from __future__ import annotations

import sys
from typing import TYPE_CHECKING
from unittest.mock import Mock, call

import pytest
import trio
from trio.testing import wait_all_tasks_blocked

from streamlink.compat import ExceptionGroup
from streamlink.utils.processoutput import ProcessOutput


if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable

    from trio.testing import MockClock


TIME_TEST_MAX = 10

# language=python
CODE = """
import _codecs
import signal
import sys

if sys.argv[-1] == "ignoresigterm":
    signal.signal(signal.SIGTERM, signal.SIG_IGN)

while line := sys.stdin.buffer.readline():
    if line[:5] == b"exit:":
        raise SystemExit(int(line[5:-1]))
    stream = sys.stdout.buffer if line[:7] == b"stdout:" else sys.stderr.buffer
    line, _ = _codecs.escape_decode(line[7:-1])
    stream.write(line)
    stream.flush()
""".strip()


class FakeProcessOutput(ProcessOutput):
    onexit: Mock
    onstdout: Mock
    onstderr: Mock

    onoutput_sender: trio.MemorySendChannel[tuple[str, str]]
    onoutput_receiver: trio.MemoryReceiveChannel[tuple[str, str]]

    def __init__(self, *args, ignoresigterm: bool = False, **kwargs):
        command = [sys.executable, "-c", CODE]
        if ignoresigterm:
            command.append("ignoresigterm")
        kwargs.setdefault("command", command)

        super().__init__(*args, **kwargs)
        self.stream_receive_some_event = trio.Event()
        self.onoutput_sender, self.onoutput_receiver = trio.open_memory_channel(10)

        def onstdouterr(channel, meth):
            def _inner(*_args, **_kwargs):
                res = meth(*_args, **_kwargs)
                try:
                    return res
                finally:
                    self.onoutput_sender.send_nowait((channel, res))

            return _inner

        self.onexit = Mock(side_effect=self.onexit)
        self.onstdout = Mock(side_effect=onstdouterr("stdout", self.onstdout))
        self.onstderr = Mock(side_effect=onstdouterr("stderr", self.onstderr))

    async def _onoutput(self, callback: Callable[[int, str], bool | None], stream: trio.abc.ReceiveStream):
        async def stream_receive_some(*_args, **_kwargs):
            self.stream_receive_some_event.set()
            return await receive_some(*_args, **_kwargs)

        receive_some = stream.receive_some
        stream.receive_some = stream_receive_some  # type: ignore

        return await super()._onoutput(callback, stream)


@pytest.fixture()
async def _max_test_time(nursery: trio.Nursery, mock_clock: MockClock):  # noqa: RUF029
    async def timeout():  # pragma: no cover
        await trio.sleep(TIME_TEST_MAX)
        mock_clock.autojump_threshold = 0
        pytest.fail("Test timed out")

    nursery.start_soon(timeout)


@pytest.fixture()
def get_process(monkeypatch: pytest.MonkeyPatch, _max_test_time) -> Callable[[], Awaitable[trio.Process]]:
    trio_run_process = trio.run_process

    # use a memory channel, so we can wait until the process has launched
    sender: trio.MemorySendChannel[trio.Process]
    receiver: trio.MemoryReceiveChannel[trio.Process]
    sender, receiver = trio.open_memory_channel(1)

    async def get_process() -> trio.Process:
        return await receiver.receive()

    async def fake_trio_run_process(*args, task_status, **kwargs):
        task_status_started = task_status.started

        def fake_task_status_started(process: trio.Process):
            task_status_started(process)
            sender.send_nowait(process)

        # intercept the task status report
        task_status.started = fake_task_status_started

        return await trio_run_process(*args, task_status=task_status, **kwargs)

    monkeypatch.setattr("trio.run_process", fake_trio_run_process)

    return get_process


@pytest.mark.trio()
@pytest.mark.usefixtures("_max_test_time")
async def test_ontimeout(mock_clock: MockClock) -> None:
    po = FakeProcessOutput(timeout=2)
    result = None

    async def run():
        nonlocal result
        result = await po.arun()

    async with trio.open_nursery() as nursery:
        nursery.start_soon(run)
        await wait_all_tasks_blocked()
        assert result is None
        mock_clock.jump(4)

    assert result is False
    assert po.onexit.call_args_list == []
    assert po.onstdout.call_args_list == []
    assert po.onstderr.call_args_list == []


@pytest.mark.trio()
@pytest.mark.parametrize(
    ("exit_code", "expected"),
    [
        pytest.param(0, True, id="success"),
        pytest.param(1, False, id="failure"),
    ],
)
async def test_exit_code(get_process: Callable[[], Awaitable[trio.Process]], exit_code: int, expected: bool):
    po = FakeProcessOutput(timeout=4)
    result = None

    async def run():
        nonlocal result
        result = await po.arun()

    async with trio.open_nursery() as nursery:
        nursery.start_soon(run)
        process = await get_process()
        assert process.stdin
        await process.stdin.send_all(f"exit:{exit_code}\n".encode())

    assert result is expected
    assert po.onexit.call_args_list == [call(exit_code)]
    assert po.onstdout.call_args_list == []
    assert po.onstderr.call_args_list == []


@pytest.mark.trio()
async def test_stdout_stderr(get_process: Callable[[], Awaitable[trio.Process]]):
    po = FakeProcessOutput(timeout=4)
    result = None

    async def run():
        nonlocal result
        result = await po.arun()

    async with trio.open_nursery() as nursery:
        nursery.start_soon(run)
        process = await get_process()
        assert process.stdin
        assert po.onstdout.call_args_list == []
        assert po.onstderr.call_args_list == []

        await process.stdin.send_all(b"stdout:foo\\n\n")
        assert await po.onoutput_receiver.receive() == ("stdout", None)
        assert po.onstdout.call_args_list == [call(0, "foo")]
        assert po.onstderr.call_args_list == []

        await process.stdin.send_all(b"stdout:bar\\nbaz\\n\n")
        assert await po.onoutput_receiver.receive() == ("stdout", None)
        assert await po.onoutput_receiver.receive() == ("stdout", None)
        assert po.onstdout.call_args_list == [call(0, "foo"), call(1, "bar"), call(2, "baz")]
        assert po.onstderr.call_args_list == []

        await process.stdin.send_all(b"stderr:bar\\n\n")
        assert await po.onoutput_receiver.receive() == ("stderr", None)
        assert po.onstdout.call_args_list == [call(0, "foo"), call(1, "bar"), call(2, "baz")]
        assert po.onstderr.call_args_list == [call(0, "bar")]

        # order of items in receive stream is not guaranteed
        await process.stdin.send_all(b"stdout:123\\n\nstderr:456\\n\n")
        received = []
        for _ in range(2):
            received.append(await po.onoutput_receiver.receive())
        received.sort(key=lambda elem: elem[0])
        assert received == [("stderr", None), ("stdout", None)]

        await process.stdin.send_all(b"exit:0\n")

    assert result is True
    assert po.onexit.call_args_list == [call(0)]
    assert po.onstdout.call_args_list == [call(0, "foo"), call(1, "bar"), call(2, "baz"), call(3, "123")]
    assert po.onstderr.call_args_list == [call(0, "bar"), call(1, "456")]


@pytest.mark.trio()
@pytest.mark.parametrize(
    ("stdin", "expected"),
    [
        pytest.param(
            (b"stdout:foo\n", b"stdout:bar\\n\n"),
            ([call(0, "foobar")], [call(0, "foobar")]),
            id="line",
        ),
        pytest.param(
            # incomplete UTF-8 sequence of the bear-face emoji, followed by its remaining data and another incomplete one
            (b"stdout:\\xf0\\x9f\n", b"stdout:\\x90\\xbb\\n\\xf0\\x9f\n"),
            # final result includes the Unicode "replacement character" due to the buffer having an incomplete UTF-8 sequence
            ([call(0, "🐻")], [call(0, "🐻"), call(1, "\ufffd")]),
            id="utf8-sequence",
        ),
    ],
)
async def test_stdout_incomplete_chunks(
    get_process: Callable[[], Awaitable[trio.Process]],
    stdin: tuple[bytes, bytes],
    expected: tuple[list, list],
):
    po = FakeProcessOutput(timeout=4)
    po._receive_max_bytes = 1  # required for full code coverage
    result = None

    async def run():
        nonlocal result
        result = await po.arun()

    async with trio.open_nursery() as nursery:
        nursery.start_soon(run)
        process = await get_process()
        assert process.stdin
        assert po.onstdout.call_args_list == []

        await process.stdin.send_all(stdin[0])
        await po.stream_receive_some_event.wait()
        assert po.onoutput_sender.statistics().current_buffer_used == 0
        assert po.onstdout.call_args_list == []

        await process.stdin.send_all(stdin[1])
        assert await po.onoutput_receiver.receive() == ("stdout", None)
        assert po.onstdout.call_args_list == expected[0]

        await process.stdin.send_all(b"exit:0\n")

    assert result is True
    assert po.onexit.call_args_list == [call(0)]
    assert po.onstdout.call_args_list == expected[1]


@pytest.mark.trio()
@pytest.mark.parametrize(
    ("expected", "return_value"),
    [
        pytest.param("yes", True, id="success"),
        pytest.param("no", False, id="failure"),
    ],
)
@pytest.mark.parametrize("channel", ["stdout", "stderr"])
async def test_output_callback(
    get_process: Callable[[], Awaitable[trio.Process]],
    expected: str,
    return_value: bool,
    channel: str,
):
    class CustomProcessOutput(ProcessOutput):
        def onstdout(self, idx: int, line: str) -> bool:
            return line == expected

        def onstderr(self, idx: int, line: str) -> bool:
            return line == expected

    class CustomFakeProcessOutput(FakeProcessOutput, CustomProcessOutput):
        pass

    po = CustomFakeProcessOutput(timeout=4)
    result = None

    async def run():
        nonlocal result
        result = await po.arun()

    async with trio.open_nursery() as nursery:
        nursery.start_soon(run)
        process = await get_process()
        assert process.stdin

        await process.stdin.send_all(f"{channel}:yes\\n\n".encode())
        assert await po.onoutput_receiver.receive() == (channel, return_value)

        await process.stdin.send_all(b"exit:0\n")

    assert result is return_value
    assert po.onexit.call_args_list == []
    assert po.onstdout.call_args_list == ([call(0, "yes")] if channel == "stdout" else [])
    assert po.onstderr.call_args_list == ([call(0, "yes")] if channel == "stderr" else [])


@pytest.mark.trio()
async def test_output_exception(get_process: Callable[[], Awaitable[trio.Process]]):
    class CustomProcessOutput(ProcessOutput):
        def onstdout(self, idx: int, line: str) -> bool:
            raise ZeroDivisionError()

    class CustomFakeProcessOutput(FakeProcessOutput, CustomProcessOutput):
        pass

    po = CustomFakeProcessOutput(timeout=4)
    # noinspection PyUnusedLocal
    process: trio.Process | None = None

    with pytest.raises(ExceptionGroup) as exc_info:  # noqa: PT012
        async with trio.open_nursery() as nursery:
            nursery.start_soon(po.arun)
            process = await get_process()
            assert process.stdin

            await process.stdin.send_all(b"stdout:foo\\n\n")
            assert await po.onoutput_receiver.receive() == ("stdout", None)

    assert process
    assert process.poll() is not None
    assert exc_info.group_contains(ZeroDivisionError)


@pytest.mark.posix_only()
@pytest.mark.trio()
async def test_kill(monkeypatch: pytest.MonkeyPatch, mock_clock: MockClock, get_process: Callable[[], Awaitable[trio.Process]]):
    po = FakeProcessOutput(timeout=2, wait_terminate=4, ignoresigterm=True)
    result = None

    async def run():
        nonlocal result
        result = await po.arun()

    async with trio.open_nursery() as nursery:
        nursery.start_soon(run)
        process = await get_process()
        assert process.poll() is None
        assert process.stdin

        # ensure that our Python subprocess has fully initialized and updated its SIGTERM signal handler
        await process.stdin.send_all(b"stdout:OK\\n\n")
        assert await po.onoutput_receiver.receive() == ("stdout", None)
        assert po.onstdout.call_args_list == [call(0, "OK")]
        po.onstdout.call_args_list.clear()

        mock_terminate = Mock(side_effect=process.terminate)
        mock_kill = Mock(side_effect=process.kill)
        monkeypatch.setattr(process, "terminate", mock_terminate)
        monkeypatch.setattr(process, "kill", mock_kill)

        mock_clock.jump(po.timeout)

        await wait_all_tasks_blocked()
        assert process.poll() is None

        assert mock_terminate.call_count == 1
        assert mock_kill.call_count == 0

        await wait_all_tasks_blocked()
        assert process.poll() is None

        mock_clock.jump(po.wait_terminate)
        await wait_all_tasks_blocked()
        assert mock_terminate.call_count == 1
        assert mock_kill.call_count == 1

        mock_clock.rate = 1

    assert result is False
    assert po.onexit.call_args_list == []
    assert po.onstdout.call_args_list == []
    assert po.onstderr.call_args_list == []



from __future__ import annotations

import sys
from contextlib import asynccontextmanager
from subprocess import PIPE
from unittest.mock import Mock

import pytest
import trio

from streamlink.webbrowser.webbrowser import Webbrowser


@pytest.fixture()
def caplog(caplog: pytest.LogCaptureFixture):
    caplog.set_level(1, "streamlink")
    return caplog


@pytest.fixture(autouse=True)
def resolve_executable(request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch):
    return_value = getattr(request, "param", "default")
    monkeypatch.setattr("streamlink.webbrowser.webbrowser.resolve_executable", Mock(return_value=return_value))
    return return_value


@pytest.fixture()
def webbrowser_launch(monkeypatch: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture):
    trio_run_process = trio.run_process

    # use a memory channel, so we can wait until the process has launched
    sender: trio.MemorySendChannel[trio.Process]
    receiver: trio.MemoryReceiveChannel[trio.Process]
    sender, receiver = trio.open_memory_channel(1)

    async def fake_trio_run_process(*args, task_status, **kwargs):
        task_status_started = task_status.started

        def fake_task_status_started(process: trio.Process):
            task_status_started(process)
            sender.send_nowait(process)

        # intercept the task status report
        task_status.started = fake_task_status_started
        # override the stdin parameter, so we can send data to our dummy process
        kwargs["stdin"] = PIPE

        return await trio_run_process(*args, task_status=task_status, **kwargs)

    monkeypatch.setattr("trio.run_process", fake_trio_run_process)

    @asynccontextmanager
    async def webbrowser_launch(*args, webbrowser: Webbrowser | None = None, **kwargs):
        # dummy web browser process, which idles until stdin receives input with an exit code
        webbrowser = webbrowser or Webbrowser()
        webbrowser.executable = sys.executable
        webbrowser.arguments = ["-c", "import sys; sys.exit(int(sys.stdin.readline()))", *webbrowser.arguments]

        headless = kwargs.get("headless", False)

        async with webbrowser.launch(*args, **kwargs) as nursery:
            assert isinstance(nursery, trio.Nursery)
            assert [(record.name, record.levelname, record.msg) for record in caplog.records] == [
                (
                    "streamlink.webbrowser.webbrowser",
                    "info",
                    f"Launching web browser: {sys.executable} ({headless=})",
                ),
            ]
            caplog.records.clear()
            # wait until the process has launched, so we can test it
            process = await receiver.receive()
            yield nursery, process

    return webbrowser_launch



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import Mock, call

import pytest

from streamlink.plugin.plugin import LOW_PRIORITY, NO_PRIORITY, NORMAL_PRIORITY
from streamlink.plugins.hls import HLSPlugin
from streamlink.stream.hls import HLSStream
from tests.plugins import PluginCanHandleUrl


if TYPE_CHECKING:
    from streamlink.session import Streamlink


class TestPluginCanHandleUrlHLSPlugin(PluginCanHandleUrl):
    __plugin__ = HLSPlugin

    should_match_groups = [
        # implicit HLS URLs
        ("example.com/foo.m3u8", {"url": "example.com/foo.m3u8"}),
        ("example.com/foo.M3U8", {"url": "example.com/foo.M3U8"}),
        ("example.com/foo.m3u8?bar", {"url": "example.com/foo.m3u8?bar"}),
        ("http://example.com/foo.m3u8", {"url": "http://example.com/foo.m3u8"}),
        ("http://example.com/foo.m3u8?bar", {"url": "http://example.com/foo.m3u8?bar"}),
        ("https://example.com/foo.m3u8", {"url": "https://example.com/foo.m3u8"}),
        ("https://example.com/foo.m3u8?bar", {"url": "https://example.com/foo.m3u8?bar"}),
        ("file://foo.m3u8", {"url": "file://foo.m3u8"}),
        ("file:///foo.m3u8", {"url": "file:///foo.m3u8"}),
        ("file://../foo.m3u8", {"url": "file://../foo.m3u8"}),
        # explicit HLS URLs with protocol prefix
        ("hls://example.com/foo?bar", {"url": "example.com/foo?bar"}),
        ("hls://http://example.com/foo?bar", {"url": "http://example.com/foo?bar"}),
        ("hls://https://example.com/foo?bar", {"url": "https://example.com/foo?bar"}),
        ("hls://file://foo", {"url": "file://foo"}),
        ("hlsvariant://example.com/foo?bar", {"url": "example.com/foo?bar"}),
        ("hlsvariant://http://example.com/foo?bar", {"url": "http://example.com/foo?bar"}),
        ("hlsvariant://https://example.com/foo?bar", {"url": "https://example.com/foo?bar"}),
        ("hlsvariant://file://foo", {"url": "file://foo"}),
        # optional parameters
        ("example.com/foo.m3u8?bar abc=def", {"url": "example.com/foo.m3u8?bar", "params": "abc=def"}),
        ("http://example.com/foo.m3u8?bar abc=def", {"url": "http://example.com/foo.m3u8?bar", "params": "abc=def"}),
        ("https://example.com/foo.m3u8?bar abc=def", {"url": "https://example.com/foo.m3u8?bar", "params": "abc=def"}),
        ("hls://https://example.com/foo?bar abc=def", {"url": "https://example.com/foo?bar", "params": "abc=def"}),
        ("hlsvariant://https://example.com/foo?bar abc=def", {"url": "https://example.com/foo?bar", "params": "abc=def"}),
    ]

    should_not_match = [
        # implicit HLS URLs must have their path end with ".m3u8"
        "example.m3u8",
        "/example.m3u8",
        "example.com/m3u8",
        "example.com/M3U8",
        "example.com/m3u8 abc=def",
        "example.com/foo.m3u8,bar",
        "example.com/foo.m3u8,bar abc=def",
        # missing parameters
        "example.com/foo.m3u8 ",
        "hls://example.com/foo ",
        "hlsvariant://example.com/foo ",
    ]


@pytest.mark.parametrize(
    ("url", "priority"),
    [
        ("http://example.com/foo.m3u8", LOW_PRIORITY),
        ("http://example.com/foo.M3U8", LOW_PRIORITY),
        ("hls://http://example.com/foo.m3u8", NORMAL_PRIORITY),
        ("hls://http://example.com/bar", NORMAL_PRIORITY),
        ("hlsvariant://http://example.com/foo.m3u8", NORMAL_PRIORITY),
        ("hlsvariant://http://example.com/bar", NORMAL_PRIORITY),
        ("http://example.com/bar", NO_PRIORITY),
    ],
)
def test_priority(url, priority):
    assert next((matcher.priority for matcher in HLSPlugin.matchers if matcher.pattern.match(url)), NO_PRIORITY) == priority


@pytest.mark.parametrize(
    ("url", "expected"),
    [
        ("example.com/foo.m3u8", "https://example.com/foo.m3u8"),
        ("http://example.com/foo.m3u8", "http://example.com/foo.m3u8"),
        ("https://example.com/foo.m3u8", "https://example.com/foo.m3u8"),
        ("file://foo.m3u8", "file://foo.m3u8"),
        ("file:///foo.m3u8", "file:///foo.m3u8"),
        ("file://../foo.m3u8", "file://../foo.m3u8"),
        ("hls://example.com/foo", "https://example.com/foo"),
        ("hls://http://example.com/foo", "http://example.com/foo"),
        ("hls://https://example.com/foo", "https://example.com/foo"),
        ("hlsvariant://example.com/foo", "https://example.com/foo"),
        ("hlsvariant://http://example.com/foo", "http://example.com/foo"),
        ("hlsvariant://https://example.com/foo", "https://example.com/foo"),
    ],
)
@pytest.mark.parametrize(
    ("isvariant", "streams"),
    [
        (False, ["live"]),
        (True, ["720p", "1080p"]),
    ],
)
def test_get_streams(
    monkeypatch: pytest.MonkeyPatch,
    session: Streamlink,
    url: str,
    expected: str,
    isvariant: bool,
    streams: list,
):
    monkeypatch.setattr("streamlink.stream.hls.HLSStream.__init__", Mock(return_value=None))

    fakestreams = {}
    if isvariant:
        fakestreams["720p"] = HLSStream(session, url)
        fakestreams["1080p"] = HLSStream(session, url)

    mock_parse_variant_playlist = Mock(return_value=fakestreams)
    monkeypatch.setattr("streamlink.stream.hls.HLSStream.parse_variant_playlist", mock_parse_variant_playlist)

    plugin = HLSPlugin(session, url)
    result = plugin.streams()
    result.pop("worst", None)
    result.pop("best", None)

    assert list(result.keys()) == streams
    assert all(isinstance(s, HLSStream) for s in result.values())
    assert mock_parse_variant_playlist.call_args_list == [call(session, expected)]



from __future__ import annotations

import os
import sys
from functools import partial
from typing import TYPE_CHECKING, Any

import pytest
import requests_mock as rm

from streamlink.session import Streamlink

# noinspection PyProtectedMember
from streamlink.utils.thread import _threadname_counters  # noqa: PLC2701


if TYPE_CHECKING:
    from collections.abc import Callable, Mapping


_TEST_CONDITION_MARKERS: Mapping[str, tuple[bool, str] | Callable[[Any], tuple[bool, str]]] = {
    "linux_only": (sys.platform == "linux", "only applicable on Linux"),
    "darwin_only": (sys.platform == "darwin", "only applicable on macOS"),
    "posix_only": (os.name == "posix", "only applicable on a POSIX OS"),
    "windows_only": (os.name == "nt", "only applicable on Windows"),
    "python": lambda *ver, **_: (  # pragma: no cover
        sys.version_info >= ver,
        f"only applicable on Python {'.'.join(str(v) for v in ver)} and above",
    ),
}

_TEST_PRIORITIES = (
    "build_backend/",
    "tests/testutils/",
    "tests/utils/",
    "tests/session/",
    None,
    "tests/stream/",
    "tests/test_plugins.py",
    "tests/plugins/",
    "tests/cli/",
)


def pytest_configure(config: pytest.Config):
    config.addinivalue_line("markers", "linux_only: tests which are only applicable on Linux")
    config.addinivalue_line("markers", "darwin_only: tests which are only applicable on macOS")
    config.addinivalue_line("markers", "posix_only: tests which are only applicable on a POSIX OS")
    config.addinivalue_line("markers", "windows_only: tests which are only applicable on Windows")
    config.addinivalue_line("markers", "python(version): tests which are only applicable on specific Python versions")
    config.addinivalue_line("markers", "nomockedhttprequest: tests where no mocked HTTP request will be made")


def pytest_runtest_setup(item: pytest.Item):
    _check_test_condition(item)


def pytest_collection_modifyitems(items: list[pytest.Item]):  # pragma: no cover
    default = next((idx for idx, string in enumerate(_TEST_PRIORITIES) if string is None), sys.maxsize)
    priorities = {
        item: next(
            (
                idx
                for idx, string in enumerate(_TEST_PRIORITIES)
                if string is not None and item.nodeid.startswith(string)
            ),
            default,
        )
        for item in items
    }  # fmt: skip
    items.sort(key=lambda item: priorities.get(item, default))


def _check_test_condition(item: pytest.Item):  # pragma: no cover
    for m in item.iter_markers():
        if m.name not in _TEST_CONDITION_MARKERS:
            continue
        data = _TEST_CONDITION_MARKERS[m.name]
        kwargs = dict(m.kwargs)
        reason = kwargs.pop("reason", None)
        if callable(data):
            assert not isinstance(data, tuple)
            cond, msg = data(*m.args, **kwargs)
        else:
            cond, msg = data
        if not cond:
            pytest.skip(msg if not reason else f"{msg} ({reason})")


# ========================
# globally shared fixtures
# ========================


@pytest.fixture()
def session(request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch):
    options = getattr(request, "param", {})
    plugins_builtin = options.pop("plugins-builtin", False)
    plugins_lazy = options.pop("plugins-lazy", False)

    session = Streamlink(
        options=options,
        plugins_builtin=plugins_builtin,
        plugins_lazy=plugins_lazy,
    )

    try:
        yield session
    finally:
        Streamlink.resolve_url.cache_clear()


@pytest.fixture()
def requests_mock(requests_mock: rm.Mocker) -> rm.Mocker:
    """
    Override of the default `requests_mock` fixture, with `InvalidRequest` raised on unknown requests
    """
    requests_mock.register_uri(rm.ANY, rm.ANY, exc=rm.exceptions.InvalidRequest)
    return requests_mock


@pytest.fixture()
def os_environ(request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch) -> dict[str, str]:
    class FakeEnviron(dict):
        def __setitem__(self, key, value):
            if key == "PYTEST_CURRENT_TEST":
                return
            return super().__setitem__(key, value)

    fakeenviron = FakeEnviron(getattr(request, "param", {}))
    monkeypatch.setattr("os.environ", fakeenviron)

    return fakeenviron


@pytest.fixture(autouse=True, scope="session")
def _patch_trio_run():
    import trio  # noqa: PLC0415

    trio_run = trio.run
    # `strict_exception_groups` changed from False to True in `trio==0.25`:
    # Patch `trio.run()` and make older versions of trio behave like `trio>=0.25`
    # as pytest-trio doesn't allow setting custom `trio.run()` args/kwargs
    trio.run = partial(trio.run, strict_exception_groups=True)  # type: ignore
    yield
    trio.run = trio_run


@pytest.fixture(autouse=True)
def _clear_threadname_counters():
    yield
    _threadname_counters.clear()



from contextlib import contextmanager
from io import BytesIO
from pathlib import Path

from lxml.etree import iterparse


__HERE__ = Path(__file__).parent.absolute()


def _parse_xml(data, strip_ns=False):
    data = bytes(data, "utf8")
    try:
        it = iterparse(BytesIO(data))
        for _, el in it:
            if "}" in el.tag and strip_ns:  # pragma: no branch
                # strip all namespaces
                el.tag = el.tag.split("}", 1)[1]
        return it.root
    except Exception as err:  # pragma: no cover
        snippet = repr(data)
        if len(snippet) > 35:
            snippet = f"{snippet[:35]} ..."

        raise ValueError(f"Unable to parse XML: {err} ({snippet})") from err


@contextmanager
def text(path, encoding="utf8"):
    with __HERE__.joinpath(path).open("r", encoding=encoding) as resource_fh:
        yield resource_fh


@contextmanager
def xml(path, encoding="utf8"):
    with __HERE__.joinpath(path).open("r", encoding=encoding) as resource_fh:
        yield _parse_xml(resource_fh.read(), strip_ns=True)



from __future__ import annotations

from argparse import Namespace
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import Mock, call

import pytest

import tests.resources
from streamlink.exceptions import NoPluginError
from streamlink_cli.main import setup_config_args


if TYPE_CHECKING:
    from streamlink import Streamlink


CONFIGDIR = Path(tests.resources.__path__[0], "cli", "config")


@pytest.fixture()
def _args(request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr("streamlink_cli.main.args", Namespace(**getattr(request, "param", {})))


@pytest.fixture()
def _config_files(request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr("streamlink_cli.main.CONFIG_FILES", getattr(request, "param", []))


@pytest.fixture()
def setup_args(monkeypatch: pytest.MonkeyPatch):
    mock_setup_args = Mock()
    monkeypatch.setattr("streamlink_cli.main.setup_args", mock_setup_args)

    return mock_setup_args


@pytest.fixture()
def session(monkeypatch: pytest.MonkeyPatch, session: Streamlink):
    def resolve_url(name):
        if name == "noplugin":
            raise NoPluginError()
        return name, Mock(__module__=name), name

    monkeypatch.setattr(session, "resolve_url", resolve_url)

    return session


# noinspection PyTestParametrized
@pytest.mark.usefixtures("_args", "_config_files")
@pytest.mark.parametrize(
    ("_args", "_config_files", "expected"),
    [
        pytest.param(
            {
                "no_config": False,
                "config": None,
                "url": None,
            },
            [
                CONFIGDIR / "primary",
            ],
            [
                CONFIGDIR / "primary",
            ],
            id="No URL, default config",
        ),
        pytest.param(
            {
                "no_config": False,
                "config": [
                    str(CONFIGDIR / "custom"),
                ],
                "url": None,
            },
            [
                CONFIGDIR / "primary",
            ],
            [
                CONFIGDIR / "custom",
            ],
            id="No URL, custom config",
        ),
        pytest.param(
            {
                "no_config": False,
                "config": [],
                "url": None,
            },
            [
                CONFIGDIR / "non-existent",
            ],
            None,
            id="No URL, non-existent default config",
        ),
        pytest.param(
            {
                "no_config": False,
                "config": [
                    str(CONFIGDIR / "non-existent"),
                ],
                "url": None,
            },
            [
                CONFIGDIR / "primary",
            ],
            None,
            id="No URL, non-existent custom config",
        ),
        pytest.param(
            {
                "no_config": False,
                "config": None,
                "url": "noplugin",
            },
            [
                CONFIGDIR / "primary",
            ],
            [
                CONFIGDIR / "primary",
            ],
            id="No plugin, default config",
        ),
        pytest.param(
            {
                "no_config": False,
                "config": [
                    str(CONFIGDIR / "custom"),
                ],
                "url": "noplugin",
            },
            [
                CONFIGDIR / "primary",
            ],
            [
                CONFIGDIR / "custom",
            ],
            id="No plugin, custom config",
        ),
        pytest.param(
            {
                "no_config": False,
                "config": [],
                "url": "noplugin",
            },
            [
                CONFIGDIR / "non-existent",
            ],
            None,
            id="No plugin, non-existent default config",
        ),
        pytest.param(
            {
                "no_config": False,
                "config": [
                    str(CONFIGDIR / "non-existent"),
                ],
                "url": "noplugin",
            },
            [
                CONFIGDIR / "primary",
            ],
            None,
            id="No plugin, non-existent custom config",
        ),
        pytest.param(
            {
                "no_config": False,
                "config": None,
                "url": "testplugin",
            },
            [
                CONFIGDIR / "primary",
            ],
            [
                CONFIGDIR / "primary",
                CONFIGDIR / "primary.testplugin",
            ],
            id="Testplugin, default config",
        ),
        pytest.param(
            {
                "no_config": False,
                "config": None,
                "url": "testplugin",
            },
            [
                CONFIGDIR / "non-existent",
            ],
            None,
            id="Testplugin, non-existent default config",
        ),
        pytest.param(
            {
                "no_config": False,
                "config": [
                    str(CONFIGDIR / "custom"),
                ],
                "url": "testplugin",
            },
            [
                CONFIGDIR / "primary",
            ],
            [
                CONFIGDIR / "custom",
                CONFIGDIR / "primary.testplugin",
            ],
            id="Testplugin, custom config",
        ),
        pytest.param(
            {
                "no_config": False,
                "config": [
                    str(CONFIGDIR / "non-existent"),
                ],
                "url": "testplugin",
            },
            [
                CONFIGDIR / "primary",
            ],
            [
                CONFIGDIR / "primary.testplugin",
            ],
            id="Testplugin, non-existent custom config",
        ),
        pytest.param(
            {
                "no_config": False,
                "config": [
                    str(CONFIGDIR / "non-existent"),
                    str(CONFIGDIR / "primary"),
                    str(CONFIGDIR / "secondary"),
                ],
                "url": "testplugin",
            },
            [
                CONFIGDIR / "primary",
            ],
            [
                CONFIGDIR / "secondary",
                CONFIGDIR / "primary",
                CONFIGDIR / "primary.testplugin",
            ],
            id="Testplugin, multiple custom configs",
        ),
        pytest.param(
            {
                "no_config": True,
                "config": [],
                "url": "testplugin",
            },
            [
                CONFIGDIR / "primary",
            ],
            None,
            id="No config, default config",
        ),
        pytest.param(
            {
                "no_config": True,
                "config": [
                    str(CONFIGDIR / "primary"),
                    str(CONFIGDIR / "secondary"),
                ],
                "url": "testplugin",
            },
            [
                CONFIGDIR / "primary",
            ],
            None,
            id="No config, multiple custom configs",
        ),
    ],
    indirect=["_args", "_config_files"],
)
@pytest.mark.parametrize("ignore_unknown", [True, False])
def test_setup_config_args(
    recwarn: pytest.WarningsRecorder,
    setup_args: Mock,
    expected: list | None,
    ignore_unknown: bool,
):
    parser = Mock()
    setup_config_args(parser, ignore_unknown=ignore_unknown)
    if expected is not None:
        assert setup_args.call_args_list == [call(parser, expected, ignore_unknown=ignore_unknown)]
    else:
        assert setup_args.call_args_list == []



from streamlink.plugins.mrtmk import MRTmk
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlMRTmk(PluginCanHandleUrl):
    __plugin__ = MRTmk

    should_match = [
        "http://play.mrt.com.mk/live/658323455489957",
        "http://play.mrt.com.mk/live/47",
        "http://play.mrt.com.mk/play/1581",
    ]

    should_not_match = [
        "http://play.mrt.com.mk/",
        "http://play.mrt.com.mk/c/2",
    ]



from __future__ import annotations

import argparse
import ast
import json
import re
import sys
from dataclasses import asdict, dataclass, is_dataclass
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING, Any, ClassVar, Generic, TextIO, TypeAlias, TypeVar


if TYPE_CHECKING:
    TListOfConstants: TypeAlias = list[bool | int | float | str | None]
    TConstantOrListOfConstants: TypeAlias = bool | int | float | str | TListOfConstants | None
    TMappingOfConstantOrListOfConstants: TypeAlias = dict[str, TConstantOrListOfConstants]


DEFAULT_PLUGINSPATH = Path(__file__).parents[1] / "src" / "streamlink" / "plugins"

PLUGINSJSON_COMMENTS = [
    "DO NOT MODIFY!",
    "This file was auto-generated and its checksum is validated before loading.",
    "If you want to modify existing plugins, then please see the plugin-sideloading or developing docs:",
    "https://streamlink.github.io/",
]


class ParseError(ValueError):
    def __init__(self, message: str, node: ast.AST):
        super().__init__(message)
        self.lineno = getattr(node, "lineno", None)
        self.col_offset = getattr(node, "col_offset", None)


@dataclass
class PluginMatcher:
    NO_PRIORITY: ClassVar[int] = 0
    LOW_PRIORITY: ClassVar[int] = 10
    NORMAL_PRIORITY: ClassVar[int] = 20
    HIGH_PRIORITY: ClassVar[int] = 30

    pattern: str
    flags: int | None = None
    priority: int | None = None
    name: str | None = None


@dataclass
class PluginArgument:
    name: str
    action: str | None = None
    nargs: int | str | None = None
    const: TConstantOrListOfConstants = None
    default: TConstantOrListOfConstants = None
    type: str | None = None
    type_args: TListOfConstants | None = None
    type_kwargs: TMappingOfConstantOrListOfConstants | None = None
    choices: TListOfConstants | None = None
    required: bool | None = None
    help: str | None = None
    metavar: str | list[str] | None = None
    dest: str | None = None
    requires: str | list[str] | None = None
    prompt: str | None = None
    sensitive: bool | None = None
    argument_name: str | None = None


@dataclass
class Plugin:
    matchers: list[PluginMatcher]
    arguments: list[PluginArgument]


_TParseResult = TypeVar("_TParseResult")


class Parser(ast.NodeVisitor, Generic[_TParseResult]):
    def visit(self, node: ast.AST) -> _TParseResult:
        return super().visit(node)


class ParseCall(ast.NodeVisitor):
    _PARSERS: ClassVar[dict[str, type[ast.NodeVisitor]]] = {}

    def visit_Call(self, node: ast.Call) -> dict[str, Any]:
        parsers = self._PARSERS

        data = {}
        parsers_list = list(parsers.items())

        for idx, arg in enumerate(node.args or []):
            if idx >= len(parsers_list):
                raise ParseError("Invalid number of arguments", arg)
            key, parser = parsers_list[idx]
            data[key] = parser().visit(arg)

        for kw in node.keywords:
            if kw.arg not in parsers:
                raise ParseError("Invalid keyword argument", kw)
            parser = parsers[kw.arg]
            data[kw.arg] = parser().visit(kw.value)

        return data


class ParseConstant(ast.NodeVisitor):
    NAME: str = "constant"
    TYPE: type | tuple[type, ...] = str
    REQUIRED: bool = False

    def generic_visit(self, node: ast.AST):
        raise ParseError(f"Invalid {self.NAME}: unknown AST node", node)

    def visit_Constant(self, node: ast.Constant):
        if (self.REQUIRED or node.value is not None) and not isinstance(node.value, self.TYPE):
            raise ParseError(f"Invalid {self.NAME} type", node)

        return node.value


class ParseConstantOrSequenceOfConstants(ParseConstant):
    def __init__(self):
        super().__init__()
        self._sequence = False

    def _visit_sequence(self, node: ast.List | ast.Tuple):
        if self._sequence:
            raise ParseError(f"Invalid {self.NAME} type", node)

        self._sequence = True

        return [self.visit(item) for item in node.elts]

    def visit_List(self, node: ast.List):
        return self._visit_sequence(node)

    def visit_Tuple(self, node: ast.Tuple):
        return self._visit_sequence(node)


class ParseSequenceOfConstants(ParseConstantOrSequenceOfConstants):
    def visit_Constant(self, node: ast.Constant):
        if (self.REQUIRED or node.value is not None) and not self._sequence:
            raise ParseError(f"Invalid {self.NAME} type", node)

        return super().visit_Constant(node)


class ParseMappingOfConstants(ParseConstantOrSequenceOfConstants):
    def __init__(self):
        super().__init__()
        self._mapping = False

    def _visit_sequence(self, node: ast.List | ast.Tuple):
        if not self._mapping:
            raise ParseError(f"Invalid {self.NAME} type", node)

        return super()._visit_sequence(node)

    def visit_Constant(self, node: ast.Constant):
        if (self.REQUIRED or node.value is not None) and not self._mapping:
            raise ParseError(f"Invalid {self.NAME} type", node)

        return super().visit_Constant(node)

    def visit_Dict(self, node: ast.Dict):
        if self._mapping:
            raise ParseError(f"Invalid {self.NAME} type", node)

        self._mapping = True

        return {self.visit(k): self.visit(v) for k, v in zip(node.keys, node.values, strict=True) if k}


class ParseBinOpOr(ast.NodeVisitor):
    def generic_visit(self, node: ast.AST):
        raise ParseError("Unsupported node type", node)

    def visit_BinOp(self, node: ast.BinOp) -> int:
        if not isinstance(node.op, ast.BitOr):
            raise ParseError("Unsupported binary operator", node)

        return self.visit(node.left) | self.visit(node.right)


class ParsePluginMatcher(ParseCall, Parser[PluginMatcher]):
    class ParseReCompile(ParseCall):
        _RE_VERBOSE_PATTERN = re.compile(r"\n+\s{2,}")

        class ParsePattern(ParseConstant):
            NAME = "@pluginmatcher pattern"
            REQUIRED = True

        class ParseFlags(ParseBinOpOr):
            _ATTRIBUTES: ClassVar[dict[str, int]] = {
                "I": re.RegexFlag.IGNORECASE,
                "IGNORECASE": re.RegexFlag.IGNORECASE,
                "S": re.RegexFlag.DOTALL,
                "DOTALL": re.RegexFlag.DOTALL,
                "X": re.RegexFlag.VERBOSE,
                "VERBOSE": re.RegexFlag.VERBOSE,
            }

            def visit_Attribute(self, node: ast.Attribute) -> int:
                if not isinstance(node.value, ast.Name) or node.value.id != "re" or node.attr not in self._ATTRIBUTES:
                    raise ParseError("Invalid attribute", node)

                return self._ATTRIBUTES[node.attr]

        _PARSERS: ClassVar[dict[str, type[ast.NodeVisitor]]] = {
            "pattern": ParsePattern,
            "flags": ParseFlags,
        }

        def generic_visit(self, node: ast.AST):
            raise ParseError("Invalid @pluginmatcher pattern: unknown AST node", node)

        def visit_Call(self, node: ast.Call):
            func = node.func
            if (
                not isinstance(func, ast.Attribute)
                or func.attr != "compile"
                or not isinstance(func.value, ast.Name)
                or func.value.id != "re"
            ):
                raise ParseError("Invalid @pluginmatcher pattern: not a compiled regex", func)

            data = super().visit_Call(node)
            pattern = data.get("pattern", "")
            flags = data.get("flags", 0)

            if not pattern:
                raise ParseError("Invalid @pluginmatcher pattern: missing pattern", node)

            if flags & re.RegexFlag.VERBOSE:
                pattern = self._RE_VERBOSE_PATTERN.sub("", pattern.strip())

            return pattern, flags

    class ParsePriority(ParseConstant, ParseBinOpOr):
        NAME = "@pluginmatcher priority"
        TYPE = int

        _ATTRIBUTES: ClassVar[dict[str, int]] = {
            "NO_PRIORITY": PluginMatcher.NO_PRIORITY,
            "LOW_PRIORITY": PluginMatcher.LOW_PRIORITY,
            "NORMAL_PRIORITY": PluginMatcher.NORMAL_PRIORITY,
            "HIGH_PRIORITY": PluginMatcher.HIGH_PRIORITY,
        }

        def visit_Name(self, node: ast.Name) -> int:
            if node.id not in self._ATTRIBUTES:
                raise ParseError("Unknown @pluginmatcher priority name", node)

            return self._ATTRIBUTES[node.id]

    class ParseName(ParseConstant):
        NAME = "@pluginmatcher name"

    _PARSERS: ClassVar[dict[str, type[ast.NodeVisitor]]] = {
        "pattern": ParseReCompile,
        "priority": ParsePriority,
        "name": ParseName,
    }

    def visit_Call(self, node: ast.Call):
        data = super().visit_Call(node)
        pattern, flags = data.get("pattern", ("", 0))
        priority = data.get("priority", PluginMatcher.priority)
        name = data.get("name", PluginMatcher.name)

        if not pattern:
            raise ParseError("Missing @pluginmatcher pattern", node)

        return PluginMatcher(
            pattern=pattern,
            flags=(None if flags == 0 else flags),
            priority=priority,
            name=name,
        )


class ParsePluginArgument(ParseCall, Parser[PluginArgument]):
    class ParseName(ParseConstant):
        NAME = "@pluginargument name"
        REQUIRED = True

    class ParseAction(ParseConstant):
        NAME = "@pluginargument action"

    class ParseNArgs(ParseConstant):
        NAME = "@pluginargument nargs"
        TYPE = int, str

    class ParseConst(ParseConstantOrSequenceOfConstants):
        NAME = "@pluginargument const"
        TYPE = bool, int, float, str

    class ParseDefault(ParseConstantOrSequenceOfConstants):
        NAME = "@pluginargument default"
        TYPE = bool, int, float, str

    class ParseType(ParseConstant):
        NAME = "@pluginargument type"

    class ParseTypeArgs(ParseSequenceOfConstants):
        NAME = "@pluginargument type_args"
        TYPE = bool, int, float, str

    class ParseTypeKwargs(ParseMappingOfConstants):
        NAME = "@pluginargument type_kwargs"
        TYPE = bool, int, float, str

    class ParseChoices(ParseSequenceOfConstants):
        NAME = "@pluginargument choices"
        TYPE = bool, int, float, str

    class ParseRequired(ParseConstant):
        NAME = "@pluginargument required"
        TYPE = bool

    class ParseHelp(ParseConstant):
        NAME = "@pluginargument help"
        REQUIRED = True

        def visit_Attribute(self, node: ast.Attribute):
            if node.value.id != "argparse" or node.attr != "SUPPRESS":  # type: ignore
                raise ParseError("Invalid @pluginargument help type", node)

            return argparse.SUPPRESS

        def visit_Constant(self, node: ast.Constant):
            return dedent(super().visit_Constant(node)).strip()

    class ParseMetavar(ParseConstantOrSequenceOfConstants):
        NAME = "@pluginargument metavar"

    class ParseDest(ParseConstant):
        NAME = "@pluginargument dest"

    class ParseRequires(ParseConstantOrSequenceOfConstants):
        NAME = "@pluginargument requires"

    class ParsePrompt(ParseConstant):
        NAME = "@pluginargument prompt"

    class ParseSensitive(ParseConstant):
        NAME = "@pluginargument sensitive"
        TYPE = bool

    class ParseArgumentName(ParseConstant):
        NAME = "@pluginargument argument_name"

    _PARSERS: ClassVar[dict[str, type[ast.NodeVisitor]]] = {
        "name": ParseName,
        "action": ParseAction,
        "nargs": ParseNArgs,
        "const": ParseConst,
        "default": ParseDefault,
        "type": ParseType,
        "type_args": ParseTypeArgs,
        "type_kwargs": ParseTypeKwargs,
        "choices": ParseChoices,
        "required": ParseRequired,
        "help": ParseHelp,
        "metavar": ParseMetavar,
        "dest": ParseDest,
        "requires": ParseRequires,
        "prompt": ParsePrompt,
        "sensitive": ParseSensitive,
        "argument_name": ParseArgumentName,
    }

    def visit_Call(self, node: ast.Call):
        data = super().visit_Call(node)

        if not data.get("name"):
            raise ParseError("Missing @pluginargument name", node)

        return PluginArgument(**data)


class PluginVisitor(ast.NodeVisitor):
    def __init__(self) -> None:
        super().__init__()
        self.name: str | None = None
        self.matchers: list[PluginMatcher] = []
        self.arguments: list[PluginArgument] = []
        self.exports: bool = False

    def generic_visit(self, node: ast.AST):
        pass

    def visit_Module(self, node: ast.Module):
        for body in node.body:
            self.visit(body)

    def visit_Assign(self, node: ast.Assign):
        if (  # pragma: no branch
            isinstance(node.value, ast.Name)
            and node.value.id == self.name
            and self.name is not None
        ):  # fmt: skip
            for target in node.targets:
                if isinstance(target, ast.Name) and target.id == "__plugin__":  # pragma: no branch
                    self.exports = True

    def visit_ClassDef(self, node: ast.ClassDef):
        for base in node.bases:
            if not isinstance(base, ast.Name):
                continue
            if base.id == "Plugin":
                break
        else:
            return

        self.name = node.name
        for decorator in node.decorator_list:
            if (
                not isinstance(decorator, ast.Call)
                or not isinstance(decorator.func, ast.Name)
            ):  # fmt: skip
                continue

            if decorator.func.id == "pluginmatcher":
                matcher = ParsePluginMatcher().visit(decorator)
                self.matchers.append(matcher)
            elif decorator.func.id == "pluginargument":  # pragma: no branch
                argument = ParsePluginArgument().visit(decorator)
                self.arguments.append(argument)


Output: TypeAlias = "dict[str, Plugin]"


class JSONEncoder(json.JSONEncoder):  # pragma: no cover
    @staticmethod
    def _filter_dataclass_none_value(items: list[tuple[str, Any]]) -> dict[str, Any]:
        return {key: val for key, val in items if val is not None}

    def default(self, o: Any) -> Any:
        # https://github.com/python/mypy/issues/17550
        if is_dataclass(o) and not isinstance(o, type):
            return asdict(o, dict_factory=self._filter_dataclass_none_value)

        return super().default(o)


def build(pluginsdir: Path = DEFAULT_PLUGINSPATH) -> Output:
    data: Output = {}
    for file in pluginsdir.glob("*.py"):
        name = file.name
        plugin = re.sub(r"\.py$", "", name)
        source = file.read_text(encoding="utf-8")

        tree = ast.parse(source, str(file))
        visitor = PluginVisitor()
        try:
            visitor.visit(tree)
        except ParseError as err:
            raise SyntaxError(
                err.args[0],
                (
                    name,
                    (err.lineno or 1),
                    (err.col_offset or 0) + 1,
                    source.split("\n")[(err.lineno or 1) - 1],
                ),
            ) from None

        if not visitor.exports or not visitor.matchers:  # pragma: no cover
            continue

        data[plugin] = Plugin(visitor.matchers, visitor.arguments)

    # noinspection PyTypeChecker
    return dict(sorted(data.items()))


def to_json(data: Output, fd: TextIO | None = None, comments: list[str] | None = None, pretty: bool = False) -> None:
    outputformat = {"separators": (",", ": "), "indent": 2} if pretty else {"separators": (",", ":")}
    textio: TextIO = fd or sys.stdout
    for line in PLUGINSJSON_COMMENTS if comments is None else comments:
        textio.write(f"// {line}\n")
    json.dump(data, textio, cls=JSONEncoder, **outputformat)  # type: ignore[arg-type, ty:invalid-argument-type]


if __name__ == "__main__":  # pragma: no cover

    def main():
        parser = argparse.ArgumentParser()
        parser.add_argument("dir", nargs="?", type=Path, default=DEFAULT_PLUGINSPATH)
        parser.add_argument("--no-comments", action="store_true")
        parser.add_argument("--pretty", action="store_true")
        parser.add_argument("-o", "--output", default="-", help="Output file")

        args = parser.parse_args()
        data = build(args.dir)

        options = {"pretty": args.pretty}
        if args.no_comments:
            options["comments"] = []

        if args.output == "-":
            to_json(data, **options)
        else:
            with Path(args.output).open("w", encoding="utf-8") as fd:
                to_json(data, fd, **options)

    main()



from streamlink.plugins.cdnbg import CDNBG
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlCDNBG(PluginCanHandleUrl):
    __plugin__ = CDNBG

    should_match_groups = [
        (("bgonair", "http://bgonair.bg/tvonline"), {}),
        (("bgonair", "http://bgonair.bg/tvonline/"), {}),
        (("nova", "http://www.nova.bg/live"), {}),
        (("nova", "http://nova.bg/live"), {}),
        (("bnt", "http://bnt.bg/live"), {}),
        (("bnt", "http://bnt.bg/live/bnt1"), {}),
        (("bnt", "http://bnt.bg/live/bnt2"), {}),
        (("bnt", "http://bnt.bg/live/bnt3"), {}),
        (("bnt", "http://bnt.bg/live/bnt4"), {}),
        (("bnt", "http://tv.bnt.bg/bnt1"), {}),
        (("bnt", "http://tv.bnt.bg/bnt2"), {}),
        (("bnt", "http://tv.bnt.bg/bnt3"), {}),
        (("bnt", "http://tv.bnt.bg/bnt4"), {}),
        (("mu-vi", "http://mu-vi.tv/LiveStreams/pages/Live.aspx"), {}),
        (("bloombergtv", "https://www.bloombergtv.bg/video"), {}),
        (("cdnbg", "https://i.cdn.bg/live/xfr3453g0d"), {}),
        (("armymedia", "https://armymedia.bg/%d0%bd%d0%b0-%d0%b6%d0%b8%d0%b2%d0%be/"), {}),
    ]

    should_not_match = [
        "https://www.tvevropa.com",
        "http://www.kanal3.bg/live",
        "http://inlife.bg/",
        "http://videochanel.bstv.bg",
        "http://video.bstv.bg/",
        "http://bitelevision.com/live",
        "http://mmtvmusic.com/live/",
        "http://chernomore.bg/",
    ]



import pkgutil
import re
import tokenize
from inspect import Parameter, signature
from pathlib import Path

import pytest

import streamlink.plugins
import tests.plugins
from streamlink.logger import StreamlinkLogger
from streamlink.plugin.plugin import Matcher, Plugin
from streamlink.utils.module import exec_module


plugins_path = streamlink.plugins.__path__[0]
plugintests_path = tests.plugins.__path__[0]

protocol_plugins = [
    "http",
    "hls",
    "dash",
]
plugintests_ignore = [
    "test_stream",
]

plugin_modules = [
    module_info
    for module_info in pkgutil.iter_modules([plugins_path])
    if not module_info.name.startswith("common_")
]  # fmt: skip
plugins = [module_info.name for module_info in plugin_modules]
plugins_no_protocols = [pname for pname in plugins if pname not in protocol_plugins]
plugintests = [
    re.sub(r"^test_", "", tname)
    for finder, tname, ispkg in pkgutil.iter_modules([plugintests_path])
    if tname.startswith("test_") and tname not in plugintests_ignore
]

PLUGIN_TYPES = "live", "vod", "live, vod"
PLUGIN_METADATA = "id", "author", "category", "title"

re_url = re.compile(r"^https?://")
re_metadata = re.compile(rf"^({'|'.join(re.escape(item) for item in PLUGIN_METADATA)})(\s.+)?$")


def unique(iterable):
    seen = set()
    for item in iterable:
        if item not in seen:
            seen.add(item)
            yield item


class TestPlugins:
    @pytest.fixture(scope="class", params=plugin_modules, ids=[moduleinfo.name for moduleinfo in plugin_modules])
    def plugin(self, request):
        return exec_module(request.param.module_finder, f"streamlink.plugins.{request.param.name}")

    def test_exports_plugin(self, plugin):
        assert hasattr(plugin, "__plugin__"), "Plugin module exports __plugin__"
        assert issubclass(plugin.__plugin__, Plugin), "__plugin__ is an instance of the Plugin class"

    def test_classname(self, plugin):
        classname = plugin.__plugin__.__name__
        assert classname == classname[0].upper() + classname[1:], "__plugin__ class name starts with uppercase letter"
        assert "_" not in classname, "__plugin__ class name does not contain underscores"

    def test_constructor(self, plugin):
        assert (
            plugin.__plugin__.__init__ is Plugin.__init__
            or tuple(
                (param.name, param.kind)
                for param in signature(plugin.__plugin__.__init__).parameters.values()
            ) == (
                ("self", Parameter.POSITIONAL_OR_KEYWORD),
                ("args", Parameter.VAR_POSITIONAL),
                ("kwargs", Parameter.VAR_KEYWORD),
            )
        )  # fmt: skip

    def test_matchers(self, plugin):
        pluginclass = plugin.__plugin__
        assert isinstance(pluginclass.matchers, list), "Has at a matchers list"
        assert len(pluginclass.matchers) > 0, "Has at least one matcher"
        assert all(isinstance(matcher, Matcher) for matcher in pluginclass.matchers), "Only has valid matchers"

    def test_plugin_api(self, plugin):
        pluginclass = plugin.__plugin__
        assert not hasattr(pluginclass, "can_handle_url"), "Does not implement deprecated can_handle_url(url)"
        assert not hasattr(pluginclass, "priority"), "Does not implement deprecated priority(url)"
        assert callable(pluginclass._get_streams), "Implements _get_streams()"

    def test_logger(self, plugin):
        logger = getattr(plugin, "log", None) or getattr(plugin, "logger", None)
        if logger is None:  # pragma: no cover
            return
        assert isinstance(logger, StreamlinkLogger)
        assert logger.name == plugin.__name__


class TestPluginTests:
    @pytest.mark.parametrize("plugin", plugins)
    def test_plugin_has_tests(self, plugin):
        assert plugin in plugintests, "Test module exists for plugin"

    @pytest.mark.parametrize("plugintest", plugintests)
    def test_test_has_plugin(self, plugintest):
        assert plugintest in plugins, "Plugin exists for test module"


class TestPluginMetadata:
    @pytest.fixture(scope="class")
    def metadata_keys_all(self):
        return (
            "description",
            "url",
            "type",
            "webbrowser",
            "metadata",
            "region",
            "account",
            "notes",
        )

    @pytest.fixture(scope="class")
    def metadata_keys_required(self):
        return (
            "description",
            "url",
            "type",
        )

    @pytest.fixture(scope="class")
    def metadata_keys_repeat(self):
        return (
            "url",
            "metadata",
            "notes",
        )

    @pytest.fixture(scope="class")
    def metadata_keys_no_repeat(self, metadata_keys_all, metadata_keys_repeat):
        return tuple(
            key
            for key in metadata_keys_all
            if key not in metadata_keys_repeat
        )  # fmt: skip

    @pytest.fixture(scope="class", params=plugins_no_protocols)
    def tokeninfo(self, request):
        with (Path(plugins_path) / f"{request.param}.py").open(encoding="utf-8") as handle:
            tokeninfo = next(tokenize.generate_tokens(handle.readline), None)

        assert type(tokeninfo) is tokenize.TokenInfo, "Parses the first token"
        assert tokeninfo.type == tokenize.STRING, "First token is a string"

        return tokeninfo

    @pytest.fixture(scope="class")
    def metadata_items(self, tokeninfo: tokenize.TokenInfo):
        match = re.search(r"^\"\"\"\n(?P<metadata>.+)\n\"\"\"$", tokeninfo.string, re.DOTALL)
        assert match is not None, "String is a properly formatted long string"

        lines = [
            re.search(r"^\$(?P<key>\w+) (?P<value>\S.+)$", line)
            for line in match.group("metadata").split("\n")
        ]  # fmt: skip
        assert all(lines), "All lines are properly formatted using the '$key value' format"

        return [(m.group("key"), m.group("value")) for m in lines if m]

    @pytest.fixture(scope="class")
    def metadata_keys(self, metadata_items):
        return tuple(key for key, value in metadata_items)

    @pytest.fixture(scope="class")
    def metadata_dict(self, metadata_keys_no_repeat, metadata_items):
        return {k: v for k, v in metadata_items if k in metadata_keys_no_repeat}

    def test_no_unknown(self, metadata_keys_all, metadata_keys):
        assert not any(True for key in metadata_keys if key not in metadata_keys_all), \
            "No unknown metadata keys are set"  # fmt: skip

    def test_required(self, metadata_keys_required, metadata_keys):
        assert all(True for tag in metadata_keys_required if tag in metadata_keys), \
            "All required metadata keys are set"  # fmt: skip

    def test_order(self, metadata_keys_all, metadata_keys):
        keys = tuple(key for key in metadata_keys_all if key in metadata_keys)
        assert keys == tuple(unique(metadata_keys)), \
            "All metadata keys are defined in order"  # fmt: skip
        assert tuple(reversed(keys)) == tuple(unique(reversed(metadata_keys))), \
            "All repeatable metadata keys are defined in order"  # fmt: skip

    def test_repeat(self, metadata_keys_repeat, metadata_keys, metadata_items):
        items = {key: tuple(v for k, v in metadata_items if k == key) for key in metadata_keys if key in metadata_keys_repeat}
        assert items == {key: tuple(unique(value)) for key, value in items.items()}, \
            "Repeatable keys don't have any duplicates"  # fmt: skip

    def test_no_repeat(self, metadata_keys_no_repeat, metadata_keys):
        keys = tuple(key for key in metadata_keys if key in metadata_keys_no_repeat)
        assert keys == tuple(unique(keys)), "Non-repeatable keys are set at most only once"

    def test_key_url(self, metadata_items):
        assert not any(re_url.match(val) for key, val in metadata_items if key == "url"), \
            "$url metadata values don't start with http:// or https://"  # fmt: skip

    def test_key_type(self, metadata_dict):
        assert metadata_dict.get("type") in PLUGIN_TYPES, \
            "$type metadata has the correct value"  # fmt: skip

    def test_key_metadata(self, metadata_items):
        assert all(re_metadata.match(val) for key, val in metadata_items if key == "metadata"), \
            "$metadata metadata values have the correct format"  # fmt: skip
        indexes = [PLUGIN_METADATA.index(val.split(" ")[0]) for key, val in metadata_items if key == "metadata"]
        assert [PLUGIN_METADATA[i] for i in indexes] == [PLUGIN_METADATA[i] for i in sorted(indexes)], \
            "$metadata metadata values are ordered correctly"  # fmt: skip



from streamlink.plugins.dlive import DLive
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlDLive(PluginCanHandleUrl):
    __plugin__ = DLive

    should_match_groups = [
        (("live", "https://dlive.tv/cryptokaprika"), {"channel": "cryptokaprika"}),
        (("live", "https://dlive.tv/cryptokaprika?query"), {"channel": "cryptokaprika"}),
        (("live", "https://dlive.tv/cryptokaprika#hash"), {"channel": "cryptokaprika"}),
        (("vod", "https://dlive.tv/p/countrycafebgky+oLCFcknSR"), {"video": "countrycafebgky+oLCFcknSR"}),
        (("vod", "https://dlive.tv/p/countrycafebgky+oLCFcknSR?query"), {"video": "countrycafebgky+oLCFcknSR"}),
        (("vod", "https://dlive.tv/p/countrycafebgky+oLCFcknSR#hash"), {"video": "countrycafebgky+oLCFcknSR"}),
    ]

    should_not_match = [
        "https://dlive.tv/",
    ]



from streamlink.plugins.goodgame import GoodGame
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlGoodGame(PluginCanHandleUrl):
    __plugin__ = GoodGame

    should_match_groups = [
        (("default", "https://goodgame.ru/CHANNELNAME"), {"channel": "CHANNELNAME"}),
        (("default", "https://goodgame.ru/CHANNELNAME/"), {"channel": "CHANNELNAME"}),
        (("default", "https://goodgame.ru/CHANNELNAME?foo=bar"), {"channel": "CHANNELNAME"}),
        (("default", "https://www.goodgame.ru/CHANNELNAME"), {"channel": "CHANNELNAME"}),
        (("player", "https://goodgame.ru/player?CHANNELID"), {"channel_id": "CHANNELID"}),
        (("player", "https://www.goodgame.ru/player?CHANNELID"), {"channel_id": "CHANNELID"}),
    ]

    should_not_match = [
        "https://goodgame.ru/player?",
    ]



from streamlink.plugins.ltv_lsm_lv import LtvLsmLv
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlLtvLsmLv(PluginCanHandleUrl):
    __plugin__ = LtvLsmLv

    should_match = [
        "https://ltv.lsm.lv/lv/tiesraide/ltv1",
        "https://ltv.lsm.lv/lv/tiesraide/ltv7",
        "https://ltv.lsm.lv/lv/tiesraide/visiem",
        "https://ltv.lsm.lv/lv/tiesraide/lr1",
        "https://ltv.lsm.lv/lv/tiesraide/lr2",
        "https://ltv.lsm.lv/lv/tiesraide/lr3",
        "https://ltv.lsm.lv/lv/tiesraide/lr4",
        "https://ltv.lsm.lv/lv/tiesraide/lr5",
        "https://ltv.lsm.lv/lv/tiesraide/lr6",
        "https://replay.lsm.lv/lv/tiesraide/ltv7/sporta-studija-aizkulises",
        "https://replay.lsm.lv/ru/efir/ltv7/sporta-studija-aizkulises",
    ]



from __future__ import annotations

import itertools
import logging
import os
import unittest
from contextlib import nullcontext
from datetime import datetime, timedelta, timezone
from threading import Event
from typing import TYPE_CHECKING, NamedTuple
from unittest.mock import Mock, call, patch

import freezegun
import pytest
import requests.adapters
import requests_mock as rm
from requests import Response
from requests.exceptions import InvalidSchema

from streamlink.exceptions import StreamlinkDeprecationWarning
from streamlink.stream.hls import (
    M3U8,
    HLSPlaylist,
    HLSSegment,
    HLSStream,
    HLSStreamReader,
    HLSStreamWorker,
    M3U8Parser,
    MuxedHLSStream,
)
from streamlink.stream.hls.hls import log
from streamlink.utils.crypto import AES, pad
from tests.mixins.stream_hls import EventedHLSStreamWorker, EventedHLSStreamWriter, Playlist, Segment, Tag, TestMixinStreamHLS
from tests.resources import text


if TYPE_CHECKING:
    from streamlink.session import Streamlink


EPOCH = datetime(2000, 1, 1, 0, 0, 0, 0, tzinfo=timezone.utc)
ONE_SECOND = timedelta(seconds=1.0)


does_not_raise = nullcontext()


class EncryptedBase:
    content: bytes
    content_plain: bytes

    def __init__(self, num, key, iv, *args, content=None, padding=b"", append=b"", **kwargs):
        super().__init__(num, *args, **kwargs)  # type: ignore
        aesCipher = AES.new(key, AES.MODE_CBC, iv)
        content = self.content if content is None else content
        padded = content + padding if padding else pad(content, AES.block_size, style="pkcs7")
        self.content_plain = content
        self.content = aesCipher.encrypt(padded) + append


class TagMap(Tag):
    def __init__(self, num, namespace, attrs=None):
        self.path = f"map{num}"
        self.content = f"[map{num}]".encode("ascii")
        super().__init__(
            "EXT-X-MAP",
            {
                "URI": self.val_quoted_string(self.url(namespace)),
                **(attrs or {}),
            },
        )


class TagMapEnc(EncryptedBase, TagMap):
    pass


class TagKey(Tag):
    _id = itertools.count()

    def __init__(self, method="NONE", uri=None, iv=None, keyformat=None, keyformatversions=None):
        attrs = {"METHOD": method}
        if uri is not False:  # pragma: no branch
            attrs["URI"] = lambda tag, namespace: tag.val_quoted_string(tag.url(namespace))
        if iv is not None:  # pragma: no branch
            attrs["IV"] = self.val_hex(iv)
        if keyformat is not None:  # pragma: no branch
            attrs["KEYFORMAT"] = self.val_quoted_string(keyformat)
        if keyformatversions is not None:  # pragma: no branch
            attrs["KEYFORMATVERSIONS"] = self.val_quoted_string(keyformatversions)
        super().__init__("EXT-X-KEY", attrs)
        self.uri = uri
        self.path = f"encryption{next(self._id)}.key"

    def url(self, namespace):
        return self.uri.format(namespace=namespace) if self.uri else super().url(namespace)


class SegmentEnc(EncryptedBase, Segment):
    pass


def test_logger_name():
    assert log.name == "streamlink.stream.hls"


def test_repr(session: Streamlink):
    stream = HLSStream(session, "https://foo.bar/playlist.m3u8")
    assert repr(stream) == "<HLSStream ['hls', 'https://foo.bar/playlist.m3u8']>"

    multivariant: M3U8[HLSSegment, HLSPlaylist] = M3U8("https://foo.bar/master.m3u8")
    multivariant.is_master = True
    stream = HLSStream(session, "https://foo.bar/playlist.m3u8", multivariant=multivariant)
    assert repr(stream) == "<HLSStream ['hls', 'https://foo.bar/playlist.m3u8', 'https://foo.bar/master.m3u8']>"


class TestHLSVariantPlaylist:
    @pytest.fixture()
    def streams(self, request: pytest.FixtureRequest, requests_mock: rm.Mocker, session: Streamlink):
        params = getattr(request, "param", {})
        base = f"http://mocked/{request.node.originalname}"

        multivariant_url = f"{base}/multivariant.m3u8"
        multivariant_playlist = params.pop("multivariant")
        assert multivariant_playlist
        with text(multivariant_playlist) as fd:
            requests_mock.get(multivariant_url, text=fd.read())

        for media_name, media_params in params.pop("media", {}).items():
            media_playlist = media_params.pop("file")
            assert media_playlist
            with text(media_playlist) as fd:
                requests_mock.get(f"{base}/{media_name}", text=fd.read(), **media_params)

        for segment_name, segment_params in params.pop("segments", {}).items():
            requests_mock.request(method=rm.ANY, url=f"{base}/{segment_name}", **segment_params)

        return HLSStream.parse_variant_playlist(session, multivariant_url, **params)

    @pytest.mark.parametrize("streams", [{"multivariant": "hls/test_master.m3u8"}], indirect=True)
    def test_variant_playlist(self, request: pytest.FixtureRequest, streams: dict[str, HLSStream]):
        assert list(streams.keys()) == ["720p", "720p_alt", "480p", "360p", "160p", "1080p (source)", "90k"]
        assert all(isinstance(stream, HLSStream) for stream in streams.values())
        assert all(stream.multivariant is not None and stream.multivariant.is_master for stream in streams.values())

        base = f"http://mocked/{request.node.originalname}"
        stream = next(iter(streams.values()))
        assert repr(stream) == f"<HLSStream ['hls', '{base}/720p.m3u8', '{base}/multivariant.m3u8']>"

        assert stream.multivariant is not None
        assert stream.multivariant.uri == f"{base}/multivariant.m3u8"

    @pytest.mark.parametrize("streams", [{"multivariant": "hls/test_multivariant_twitch_usher_v2.m3u8"}], indirect=True)
    def test_framerate(self, streams: dict[str, HLSStream]):
        assert sorted(streams.keys()) == ["1080p60", "160k", "160p", "360p", "480p", "720p60"]

    @pytest.mark.parametrize(
        ("streams", "expected"),
        [
            pytest.param(
                {
                    "check_streams": True,
                    "multivariant": "hls/test_simple_multivariant.m3u8",
                    "media": {
                        "720p/playlist.m3u8": {"file": "hls/test_simple_media.m3u8"},
                        "1080p/playlist.m3u8": {"file": "hls/test_simple_media.m3u8"},
                    },
                    "segments": {
                        "720p/segment2.ts": {"content": b"content"},
                        "1080p/segment2.ts": {"content": b"content"},
                    },
                },
                ["1080p (source)", "720p"],
                id="all-valid",
            ),
            pytest.param(
                {
                    "check_streams": True,
                    "multivariant": "hls/test_simple_multivariant.m3u8",
                    "media": {
                        "720p/playlist.m3u8": {"file": "hls/test_simple_media.m3u8"},
                    },
                },
                ["720p"],
                id="playlist-failure",
            ),
            pytest.param(
                {
                    "check_streams": True,
                    "multivariant": "hls/test_simple_multivariant.m3u8",
                    "media": {
                        "720p/playlist.m3u8": {"file": "hls/test_simple_multivariant.m3u8"},
                        "1080p/playlist.m3u8": {"file": "hls/test_simple_multivariant.m3u8"},
                    },
                },
                [],
                id="nested-multivariant",
            ),
            pytest.param(
                {
                    "check_streams": "segments",
                    "multivariant": "hls/test_simple_multivariant.m3u8",
                    "media": {
                        "720p/playlist.m3u8": {"file": "hls/test_simple_media.m3u8"},
                        "1080p/playlist.m3u8": {"file": "hls/test_simple_media.m3u8"},
                    },
                    "segments": {
                        "720p/segment1.ts": {"content": b"content"},
                    },
                },
                ["1080p (source)", "720p"],
                id="segment-failure-first-success",
            ),
            pytest.param(
                {
                    "check_streams": "segments",
                    "multivariant": "hls/test_simple_multivariant.m3u8",
                    "media": {
                        "720p/playlist.m3u8": {"file": "hls/test_simple_media.m3u8"},
                        "1080p/playlist.m3u8": {"file": "hls/test_simple_media.m3u8"},
                    },
                    "segments": {
                        "1080p/segment1.ts": {"content": b"content"},
                    },
                },
                [],
                id="segment-failure-first-failure",
            ),
        ],
        indirect=["streams"],
    )
    def test_check_streams(self, streams: dict[str, HLSStream], expected: list[str]):
        assert sorted(streams.keys()) == expected


class EventedWorkerHLSStreamReader(HLSStreamReader):
    __worker__ = EventedHLSStreamWorker


class EventedWriterHLSStreamReader(HLSStreamReader):
    __writer__ = EventedHLSStreamWriter


class EventedWorkerHLSStream(HLSStream):
    __reader__ = EventedWorkerHLSStreamReader


class EventedWriterHLSStream(HLSStream):
    __reader__ = EventedWriterHLSStreamReader


@patch("streamlink.stream.hls.hls.HLSStreamWorker.wait", Mock(return_value=True))
class TestHLSStream(TestMixinStreamHLS, unittest.TestCase):
    def get_session(self, options=None, *args, **kwargs):
        session = super().get_session(options)
        session.set_option("hls-live-edge", 3)

        return session

    def test_thread_names(self):
        testid = self.id()
        self.subject(playlists=[Playlist(0, [Segment(0)], end=True)])
        assert self.thread.reader.worker.name == f"HLSStreamWorker-{testid}-0"
        assert self.thread.reader.writer.name == f"HLSStreamWriter-{testid}-0"
        assert self.thread.reader.writer.executor._thread_name_prefix == f"HLSStreamWriter-{testid}-0-executor"
        self.await_read(read_all=True)

    def test_playlist_end(self):
        segments = self.subject([
            Playlist(0, [Segment(0)], end=True),
        ])

        assert self.await_read(read_all=True) == self.content(segments), "Stream ends and read-all handshake doesn't time out"

    def test_playlist_end_on_empty_reload(self):
        segments = self.subject([
            Playlist(0, [Segment(0)]),
            Playlist(0, [Segment(0)], end=True),
        ])

        assert self.await_read(read_all=True) == self.content(segments), "Stream ends and read-all handshake doesn't time out"

    @patch("streamlink.stream.segmented.segmented.log")
    def test_duration(self, mock_log: Mock):
        segments = self.subject(
            [
                Playlist(
                    0,
                    [
                        Segment(0, duration=2.0),
                        Segment(1, duration=2.0),
                        Segment(2, duration=2.0),
                        Segment(3, duration=2.0),
                    ],
                    end=True,
                ),
            ],
            streamoptions={"duration": 5},
        )

        data = self.await_read(read_all=True)
        assert data == self.content(segments, cond=lambda s: 0 <= s.num < 3), "Respects the duration"
        assert all(self.called(s) for s in segments.values() if 0 <= s.num < 3), "Downloads first, second and third segment"
        assert not any(self.called(s) for s in segments.values() if s.num >= 3), "Skips other segments"
        assert mock_log.info.call_args_list == [call("Stopping stream early after 5.00s")]

    @patch("streamlink.stream.segmented.segmented.log")
    def test_offset_and_duration(self, mock_log: Mock):
        segments = self.subject(
            [
                Playlist(
                    0,
                    [
                        Segment(0, duration=1.0),
                        Segment(1, duration=0.5),
                        Segment(2, duration=0.5),
                        Segment(3, duration=1.0),
                    ],
                    end=True,
                ),
            ],
            streamoptions={"start_offset": 1, "duration": 1},
        )

        data = self.await_read(read_all=True)
        assert data == self.content(segments, cond=lambda s: 0 < s.num < 3), "Respects the offset and duration"
        assert all(self.called(s) for s in segments.values() if 0 < s.num < 3), "Downloads second and third segment"
        assert not any(self.called(s) for s in segments.values() if 0 > s.num > 3), "Skips other segments"
        assert mock_log.info.call_args_list == [call("Stopping stream early after 1.00s")]

    def test_map(self):
        discontinuity = Tag("EXT-X-DISCONTINUITY")
        map1 = TagMap(1, self.id())
        map2 = TagMap(2, self.id())
        self.mock("GET", self.url(map1), content=map1.content)
        self.mock("GET", self.url(map2), content=map2.content)

        segments = self.subject(
            [
                Playlist(0, [map1, Segment(0), Segment(1), Segment(2), Segment(3)]),
                Playlist(4, [map1, Segment(4), map2, Segment(5), Segment(6), discontinuity, map1, Segment(7)], end=True),
            ],
            options={"stream-segment-threads": 2},
        )

        data = self.await_read(read_all=True, timeout=None)
        assert data == self.content([
            map1, segments[1], segments[2], segments[3],
            segments[4], map2, segments[5], segments[6], map1, segments[7],
        ])  # fmt: skip
        assert self.called(map1, once=True), "Downloads first map only once"
        assert self.called(map2, once=True), "Downloads second map only once"


# TODO: finally rewrite the segmented/HLS test setup using pytest and replace redundant setups with parametrization
@patch("streamlink.stream.hls.hls.HLSStreamWorker.wait", Mock(return_value=True))
class TestCheckSequenceGap(TestMixinStreamHLS, unittest.TestCase):
    @patch("streamlink.stream.segmented.segmented.log")
    def test_no_discontinuity(self, mock_log: Mock):
        segments = self.subject([
            Playlist(0, [Segment(0), Segment(1)]),
            Playlist(2, [Segment(2), Segment(3)]),
            Playlist(4, [Segment(4), Segment(5)], end=True),
        ])

        data = self.await_read(read_all=True)
        assert data == self.content(segments)
        assert all(self.called(s) for s in segments.values())
        assert mock_log.warning.call_args_list == []

    @patch("streamlink.stream.segmented.segmented.log")
    def test_discontinuity_single_segment(self, mock_log: Mock):
        segments = self.subject([
            Playlist(0, [Segment(0), Segment(1)]),
            Playlist(2, [Segment(2), Segment(3)]),
            Playlist(5, [Segment(5), Segment(6)]),
            Playlist(8, [Segment(8), Segment(9)], end=True),
        ])

        data = self.await_read(read_all=True)
        assert data == self.content(segments)
        assert all(self.called(s) for s in segments.values())
        assert mock_log.warning.call_args_list == [
            call("Sequence gap of 1 segment at position 4. This is unsupported and will result in incoherent output data."),
            call("Sequence gap of 1 segment at position 7. This is unsupported and will result in incoherent output data."),
        ]

    @patch("streamlink.stream.segmented.segmented.log")
    def test_discontinuity_multiple_segments(self, mock_log: Mock):
        segments = self.subject([
            Playlist(0, [Segment(0), Segment(1)]),
            Playlist(2, [Segment(2), Segment(3)]),
            Playlist(6, [Segment(6), Segment(7)]),
            Playlist(10, [Segment(10), Segment(11)], end=True),
        ])

        data = self.await_read(read_all=True)
        assert data == self.content(segments)
        assert all(self.called(s) for s in segments.values())
        assert mock_log.warning.call_args_list == [
            call("Sequence gap of 2 segments at position 4. This is unsupported and will result in incoherent output data."),
            call("Sequence gap of 2 segments at position 8. This is unsupported and will result in incoherent output data."),
        ]


class TestHLSStreamWorker(TestMixinStreamHLS, unittest.TestCase):
    __stream__ = EventedWorkerHLSStream

    OPTIONS = {"stream-timeout": 1}

    def tearDown(self) -> None:
        worker: EventedHLSStreamWorker = self.thread.reader.worker  # type: ignore[assignment, ty:invalid-assignment]
        # don't await the handshakes on error
        worker.handshake_wait.go()
        worker.handshake_reload.go()
        return super().tearDown()

    def get_session(self, options=None, *args, **kwargs):
        return super().get_session({**self.OPTIONS, **(options or {})}, *args, **kwargs)

    def test_queue_deadline_reached(self) -> None:
        self.subject(
            start=False,
            playlists=[
                Playlist(0, targetduration=5, segments=[Segment(0)]),
                # no EXT-X-ENDLIST, last mocked playlist response will be repreated forever
                Playlist(0, targetduration=5, segments=[Segment(0), Segment(1)]),
            ],
        )
        worker: EventedHLSStreamWorker = self.thread.reader.worker  # type: ignore[assignment, ty:invalid-assignment]
        targetduration = ONE_SECOND * 5

        with (
            freezegun.freeze_time(EPOCH) as frozen_time,
            patch("streamlink.stream.segmented.segmented.log") as mock_log,
        ):
            self.start()

            assert worker.handshake_reload.wait_ready(1), "Loads playlist for the first time"
            assert worker.sequence == -1, "Initial sequence number"
            assert worker._queue_last == EPOCH, "Sets the initial last queue time"

            # first playlist reload has taken one second
            frozen_time.tick(ONE_SECOND)
            self.await_reload(1)

            assert worker.handshake_wait.wait_ready(1), "Arrives at wait() call #1"
            assert worker.sequence == 1, "Updates the sequence number"
            assert worker._queue_last == EPOCH + ONE_SECOND, "Updates the last queue time"
            assert worker.playlist_targetduration == pytest.approx(5.0)

            # trigger next reload when the target duration has passed
            frozen_time.tick(targetduration)
            self.await_playlist_wait(1)
            self.await_reload(1)

            assert worker.handshake_wait.wait_ready(1), "Arrives at wait() call #2"
            assert worker.sequence == 2, "Updates the sequence number again"
            assert worker._queue_last == EPOCH + ONE_SECOND + targetduration, "Updates the last queue time again"
            assert worker.playlist_targetduration == pytest.approx(5.0)

            for num in range(3, 6):
                # trigger next reload when the target duration has passed
                frozen_time.tick(targetduration)
                self.await_playlist_wait(1)
                self.await_reload(1)

                assert worker.handshake_wait.wait_ready(1), f"Arrives at wait() call #{num}"
                assert worker.sequence == 2, "Sequence number is unchanged"
                assert worker._queue_last == EPOCH + ONE_SECOND + targetduration, "Last queue time is unchanged"
                assert worker.playlist_targetduration == pytest.approx(5.0)

            assert mock_log.warning.call_args_list == []

            # trigger next reload when the target duration has passed
            frozen_time.tick(targetduration)
            self.await_playlist_wait(1)
            self.await_reload(1)

            self.await_read(read_all=True)
            self.await_close(1)

            assert mock_log.warning.call_args_list == [call("No new segments for more than 15.00s. Stopping...")]

    def test_queue_deadline_reached_ignored(self) -> None:
        segments = self.subject(
            start=False,
            options={"stream-segmented-queue-deadline": 0.0},
            playlists=[
                # no EXT-X-ENDLIST, last mocked playlist response will be repreated forever
                Playlist(0, targetduration=5, segments=[Segment(0)]),
            ],
        )
        worker: EventedHLSStreamWorker = self.thread.reader.worker  # type: ignore[assignment, ty:invalid-assignment]
        targetduration = ONE_SECOND * 5

        with freezegun.freeze_time(EPOCH) as frozen_time:
            self.start()

            assert worker.handshake_reload.wait_ready(1), "Loads playlist for the first time"
            assert worker.sequence == -1, "Initial sequence number"
            assert worker._queue_last == EPOCH, "Sets the initial last queue time"

            # first playlist reload has taken one second
            frozen_time.tick(ONE_SECOND)
            self.await_reload(1)

            assert worker.handshake_wait.wait_ready(1), "Arrives at first wait() call"
            assert worker.sequence == 1, "Updates the sequence number"
            assert worker._queue_last == EPOCH + ONE_SECOND, "Updates the last queue time"
            assert worker.playlist_targetduration == pytest.approx(5.0)
            assert self.await_read() == self.content(segments)

            # keep reloading a couple of times
            for num in range(10):
                frozen_time.tick(targetduration)
                self.await_playlist_wait(1)
                self.await_reload(1)

                assert worker.handshake_wait.wait_ready(1), f"Arrives at wait() #{num + 1}"
                assert worker.sequence == 1, "Sequence number is unchanged"
                assert worker._queue_last == EPOCH + ONE_SECOND, "Last queue time is unchanged"

        assert self.thread.data == [], "No new data"
        assert worker.is_alive()

        # make stream end gracefully to avoid any unnecessary thread blocking
        self.thread.reader.writer.put(None)

    def test_queue_deadline_reached_min(self) -> None:
        self.subject(
            start=False,
            playlists=[
                # no EXT-X-ENDLIST, last mocked playlist response will be repreated forever
                Playlist(0, targetduration=1, segments=[Segment(0)]),
            ],
        )
        worker: EventedHLSStreamWorker = self.thread.reader.worker  # type: ignore[assignment, ty:invalid-assignment]
        targetduration = ONE_SECOND

        with (
            freezegun.freeze_time(EPOCH) as frozen_time,
            patch("streamlink.stream.segmented.segmented.log") as mock_log,
        ):
            self.start()

            assert worker.handshake_reload.wait_ready(1), "Loads playlist for the first time"
            assert worker.sequence == -1, "Initial sequence number"
            assert worker._queue_last == EPOCH, "Sets the initial last queue time"

            # first playlist reload has taken one second
            frozen_time.tick(ONE_SECOND)
            self.await_reload(1)

            assert worker.handshake_wait.wait_ready(1), "Arrives at wait() call #1"
            assert worker.sequence == 1, "Updates the sequence number"
            assert worker._queue_last == EPOCH + ONE_SECOND, "Updates the last queue time"
            assert worker.playlist_targetduration == pytest.approx(1.0)

            for num in range(2, 7):
                # trigger next reload when the target duration has passed
                frozen_time.tick(targetduration)
                self.await_playlist_wait(1)
                self.await_reload(1)

                assert worker.handshake_wait.wait_ready(1), f"Arrives at wait() call #{num}"
                assert worker.sequence == 1, "Sequence number is unchanged"
                assert worker._queue_last == EPOCH + ONE_SECOND, "Last queue time is unchanged"
                assert worker.playlist_targetduration == pytest.approx(1.0)

            assert mock_log.warning.call_args_list == []

            # trigger next reload when the target duration has passed
            frozen_time.tick(targetduration)
            self.await_playlist_wait(1)
            self.await_reload(1)

            # FIXME: NO-GIL
            #   Closing only the worker thread when no new segments were queued keeps the writer thread and buffer still open,
            #   which is why the test's reader thread keeps running until the test teardown,
            #   but this somehow breaks the assertion down below, so close everything manually...
            #   These tests will have to be rewritten eventually in pytest-style, without having to mock log calls.
            self.close()
            self.await_close()

            assert mock_log.warning.call_args_list == [call("No new segments for more than 5.00s. Stopping...")]

    def test_playlist_reload_offset(self) -> None:
        segments = self.subject(
            start=False,
            playlists=[
                Playlist(0, targetduration=5, segments=[Segment(0)]),
                Playlist(1, targetduration=5, segments=[Segment(1)]),
                Playlist(2, targetduration=5, segments=[Segment(2)]),
                Playlist(3, targetduration=5, segments=[Segment(3)]),
                Playlist(4, targetduration=5, segments=[Segment(4)], end=True),
            ],
        )
        worker: EventedHLSStreamWorker = self.thread.reader.worker  # type: ignore[assignment, ty:invalid-assignment]
        targetduration = ONE_SECOND * 5

        with freezegun.freeze_time(EPOCH) as frozen_time:
            self.start()

            assert worker.handshake_reload.wait_ready(1), "Arrives at initial playlist reload"
            assert worker._reload_last == EPOCH, "Sets the initial value of the last reload time"

            # adjust clock and reload playlist: let it take one second
            frozen_time.move_to(worker._reload_last + ONE_SECOND)
            self.await_reload()
            assert worker._reload_time == pytest.approx(5.0), "Uses the playlist's targetduration as reload time"

            # time_completed = 00:00:01; time_elapsed = 1s
            assert worker.handshake_wait.wait_ready(1), "Arrives at first wait() call"
            assert worker.sequence == 1, "Has queued first segment"
            assert worker.playlist_end is None, "Stream hasn't ended yet"
            assert worker.time_wait == pytest.approx(4.0), "Waits for 4 seconds out of the 5 seconds reload time"
            self.await_playlist_wait()

            assert worker.handshake_reload.wait_ready(1), "Arrives at second playlist reload"
            assert worker._reload_last == EPOCH + targetduration, \
                "Last reload time is the sum of reload+wait time (=targetduration)"  # fmt: skip

            # adjust clock and reload playlist: let it exceed targetduration by two seconds
            frozen_time.move_to(worker._reload_last + targetduration + ONE_SECOND * 2)
            self.await_reload()
            assert worker._reload_time == pytest.approx(5.0), "Uses the playlist's targetduration as reload time"

            # time_completed = 00:00:12; time_elapsed = 7s (exceeded 5s targetduration)
            assert worker.handshake_wait.wait_ready(1), "Arrives at second wait() call"
            assert worker.sequence == 2, "Has queued second segment"
            assert worker.playlist_end is None, "Stream hasn't ended yet"
            assert worker.time_wait == pytest.approx(0.0), "Doesn't wait when reloading took too long"
            self.await_playlist_wait()

            assert worker.handshake_reload.wait_ready(1), "Arrives at third playlist reload"
            assert worker._reload_last == EPOCH + targetduration * 2 + ONE_SECOND * 2, \
                "Sets last reload time to current time when reloading took too long (changes the interval)"  # fmt: skip

            # adjust clock and reload playlist: let it take one second again
            frozen_time.move_to(worker._reload_last + ONE_SECOND)
            self.await_reload()
            assert worker._reload_time == pytest.approx(5.0), "Uses the playlist's targetduration as reload time"

            # time_completed = 00:00:13; time_elapsed = 1s
            assert worker.handshake_wait.wait_ready(1), "Arrives at third wait() call"
            assert worker.sequence == 3, "Has queued third segment"
            assert worker.playlist_end is None, "Stream hasn't ended yet"
            assert worker.time_wait == pytest.approx(4.0), "Waits for 4 seconds out of the 5 seconds reload time"
            self.await_playlist_wait()

            assert worker.handshake_reload.wait_ready(1), "Arrives at fourth playlist reload"
            assert worker._reload_last == EPOCH + targetduration * 3 + ONE_SECOND * 2, \
                "Last reload time is the sum of reload+wait time (=targetduration) of the changed interval"  # fmt: skip

            # adjust clock and reload playlist: simulate no fetch+processing delay
            frozen_time.move_to(worker._reload_last)
            self.await_reload()
            assert worker._reload_time == pytest.approx(5.0), "Uses the playlist's targetduration as reload time"

            # time_completed = 00:00:17; time_elapsed = 0s
            assert worker.handshake_wait.wait_ready(1), "Arrives at fourth wait() call"
            assert worker.sequence == 4, "Has queued fourth segment"
            assert worker.playlist_end is None, "Stream hasn't ended yet"
            assert worker.time_wait == pytest.approx(5.0), "Waits for the whole reload time"
            self.await_playlist_wait()

            assert worker.handshake_reload.wait_ready(1), "Arrives at fifth playlist reload"
            assert worker._reload_last == EPOCH + targetduration * 4 + ONE_SECOND * 2, \
                "Last reload time is the sum of reload+wait time (no delay)"  # fmt: skip

            # adjusting the clock is not needed anymore
            self.await_reload()
            assert self.await_read(read_all=True) == self.content(segments)
            self.await_close()
            assert worker.playlist_end == 4, "Stream has ended"
            assert not worker.handshake_wait.wait_ready(0), "Doesn't wait once ended"
            assert not worker.handshake_reload.wait_ready(0), "Doesn't reload playlist once ended"


class TestHLSStreamWorkerOptions:
    @pytest.mark.parametrize(
        ("options", "session", "expected", "warning"),
        [
            pytest.param(
                {},
                {},
                0.0,
                [],
                id="no-duration",
            ),
            pytest.param(
                {"duration": 123.45},
                {},
                123.45,
                [],
                id="duration",
            ),
            pytest.param(
                {},
                {"stream-segmented-duration": 123.45},
                123.45,
                [],
                id="stream-segmented-duration",
            ),
            pytest.param(
                {},
                {"hls-duration": 123.45},
                123.45,
                [
                    (
                        StreamlinkDeprecationWarning,
                        "`hls-duration` has been deprecated in favor of the `stream-segmented-duration` option",
                    ),
                ],
                id="hls-duration",
            ),
            pytest.param(
                {"duration": 123.45},
                {"stream-segmented-duration": 543.21},
                123.45,
                [],
                id="duration-priority",
            ),
        ],
        indirect=["session"],
    )
    def test_duration(
        self,
        recwarn: pytest.WarningsRecorder,
        session: Streamlink,
        options: dict,
        expected: float | None,
        warning: list,
    ):
        stream = HLSStream(session, "https://foo/", **options)
        reader = HLSStreamReader(stream)
        worker = HLSStreamWorker(reader)

        assert worker.duration_limit == expected
        assert [(record.category, str(record.message)) for record in recwarn.list] == warning

    @pytest.mark.parametrize(
        ("session", "expected", "warning"),
        [
            pytest.param(
                {},
                3.0,
                [],
                id="default-value",
            ),
            pytest.param(
                {"stream-segmented-queue-deadline": 5.0},
                5.0,
                [],
                id="stream-segmented-queue-deadline",
            ),
            pytest.param(
                {"hls-segment-queue-threshold": 5.0},
                5.0,
                [
                    (
                        StreamlinkDeprecationWarning,
                        "`hls-segment-queue-threshold` has been deprecated in favor of "
                        + "the `stream-segmented-queue-deadline` option",
                    ),
                ],
                id="hls-segment-queue-threshold",
            ),
        ],
        indirect=["session"],
    )
    def test_queue_deadline(
        self,
        recwarn: pytest.WarningsRecorder,
        session: Streamlink,
        expected: float,
        warning: list,
    ):
        stream = HLSStream(session, "https://foo/")
        reader = HLSStreamReader(stream)
        worker = HLSStreamWorker(reader)

        assert worker._queue_deadline_factor == expected
        assert [(record.category, str(record.message)) for record in recwarn.list] == warning


class TestHLSStreamWorkerPlaylistSequenceWarning:
    warns = pytest.mark.parametrize(
        "_assert_warning",
        [
            pytest.param([
                (
                    StreamlinkDeprecationWarning,
                    "HLSStreamWorker.playlist_sequence has been moved to SegmentedStreamWorker.sequence",
                    __file__,
                ),
            ]),
        ],
        indirect=["_assert_warning"],
    )

    @pytest.fixture()
    def reader(self, session: Streamlink):
        stream = HLSStream(session, "")
        return HLSStreamReader(stream)

    @pytest.fixture(autouse=True)
    def _assert_warning(self, request: pytest.FixtureRequest, recwarn: pytest.WarningsRecorder):
        warnings = getattr(request, "param", [])
        yield
        assert [(record.category, str(record.message), record.filename) for record in recwarn.list] == warnings

    def test_getter(self, reader: HLSStreamReader):
        assert reader.worker.sequence == -1

    @warns
    def test_warns_getter(self, reader: HLSStreamReader):
        assert reader.worker.playlist_sequence == -1

    @warns
    def test_warns_setter(self, reader: HLSStreamReader):
        reader.worker.playlist_sequence = 0


@patch("streamlink.stream.hls.hls.HLSStreamWorker.wait", Mock(return_value=True))
class TestHLSStreamByterange(TestMixinStreamHLS, unittest.TestCase):
    __stream__ = EventedWriterHLSStream

    # The dummy segments in the error tests are required because the writer's run loop would otherwise continue forever
    # due to the segment's future result being None (no requests result), and we can't await the end of the stream
    # without waiting for the stream's timeout error. The dummy segments ensure that we can call await_write for these
    # successful segments, so we can close the stream afterward and safely make the test assertions.
    # The EventedHLSStreamWriter could also implement await_fetch, but this is unnecessarily more complex than it already is.

    @patch("streamlink.stream.hls.hls.log")
    def test_unknown_offset(self, mock_log: Mock):
        self.subject([
            Playlist(
                0,
                [
                    Tag("EXT-X-BYTERANGE", "3"),
                    Segment(0),
                    Segment(1),
                ],
                end=True,
            ),
        ])

        self.await_write(2 - 1)
        self.thread.close()

        assert mock_log.error.call_args_list == [
            call("Failed to fetch segment 0: Missing BYTERANGE offset"),
        ]
        assert not self.called(Segment(0))

    @patch("streamlink.stream.hls.hls.log")
    def test_unknown_offset_map(self, mock_log: Mock):
        map1 = TagMap(1, self.id(), {"BYTERANGE": '"1234"'})
        self.mock("GET", self.url(map1), content=map1.content)
        self.subject([
            Playlist(
                0,
                [
                    Segment(0),
                    map1,
                    Segment(1),
                ],
                end=True,
            ),
        ])

        self.await_write(3 - 1)
        self.thread.close()

        assert mock_log.error.call_args_list == [
            call("Failed to fetch map for segment 1: Missing BYTERANGE offset"),
        ]
        assert not self.called(map1)

    @patch("streamlink.stream.hls.hls.log")
    def test_invalid_offset_reference(self, mock_log: Mock):
        self.subject([
            Playlist(
                0,
                [
                    Tag("EXT-X-BYTERANGE", "3@0"),
                    Segment(0),
                    Segment(1),
                    Tag("EXT-X-BYTERANGE", "5"),
                    Segment(2),
                    Segment(3),
                ],
                end=True,
            ),
        ])

        self.await_write(4 - 1)
        self.thread.close()

        assert mock_log.error.call_args_list == [
            call("Failed to fetch segment 2: Missing BYTERANGE offset"),
        ]
        assert self.mocks[self.url(Segment(0))].last_request._request.headers["Range"] == "bytes=0-2"
        assert not self.called(Segment(2))

    def test_offsets(self):
        map1 = TagMap(1, self.id(), {"BYTERANGE": '"1234@0"'})
        map2 = TagMap(2, self.id(), {"BYTERANGE": '"42@1337"'})
        self.mock("GET", self.url(map1), content=map1.content)
        self.mock("GET", self.url(map2), content=map2.content)
        s1, s2, s3, s4, s5 = Segment(0), Segment(1), Segment(2), Segment(3), Segment(4)

        self.subject([
            Playlist(
                0,
                [
                    map1,
                    Tag("EXT-X-BYTERANGE", "5@3"),
                    s1,
                    Tag("EXT-X-BYTERANGE", "7"),
                    s2,
                    map2,
                    Tag("EXT-X-BYTERANGE", "11"),
                    s3,
                    Tag("EXT-X-BYTERANGE", "17@13"),
                    s4,
                    Tag("EXT-X-BYTERANGE", "19"),
                    s5,
                ],
                end=True,
            ),
        ])

        self.await_write(1 + 2 + 1 + 3)  # 1 map, 2 partial segments, 1 map, 3 partial segments
        self.await_read(read_all=True)
        assert self.mocks[self.url(map1)].last_request._request.headers["Range"] == "bytes=0-1233"
        assert self.mocks[self.url(map2)].last_request._request.headers["Range"] == "bytes=1337-1378"
        assert self.mocks[self.url(s1)].last_request._request.headers["Range"] == "bytes=3-7"
        assert self.mocks[self.url(s2)].last_request._request.headers["Range"] == "bytes=8-14"
        assert self.mocks[self.url(s3)].last_request._request.headers["Range"] == "bytes=15-25"
        assert self.mocks[self.url(s4)].last_request._request.headers["Range"] == "bytes=13-29"
        assert self.mocks[self.url(s5)].last_request._request.headers["Range"] == "bytes=30-48"


@patch("streamlink.stream.hls.hls.HLSStreamWorker.wait", Mock(return_value=True))
class TestHLSStreamEncrypted(TestMixinStreamHLS, unittest.TestCase):
    __stream__ = EventedWriterHLSStream

    def get_session(self, options=None, *args, **kwargs):
        options = options or {}
        options.setdefault("hls-live-edge", 3)
        options.setdefault("http-headers", {"X-FOO": "BAR"})
        session = super().get_session(options)

        return session

    def gen_key(
        self,
        aes_key=None,
        aes_iv=None,
        method="AES-128",
        uri=None,
        keyformat="identity",
        keyformatversions=1,
        mock=None,
    ):
        aes_key = aes_key or os.urandom(16)
        aes_iv = aes_iv or os.urandom(16)

        key = TagKey(method=method, uri=uri, iv=aes_iv, keyformat=keyformat, keyformatversions=keyformatversions)
        self.mock("GET", key.url(self.id()), **(mock if mock else {"content": aes_key}))

        return aes_key, aes_iv, key

    @patch("streamlink.stream.hls.hls.log")
    def test_hls_encrypted_invalid_method(self, mock_log: Mock):
        aesKey, aesIv, key = self.gen_key(method="INVALID")

        self.subject([
            Playlist(0, [key, SegmentEnc(1, aesKey, aesIv)], end=True),
        ])
        self.await_write()

        self.thread.close()
        self.await_close()

        assert b"".join(self.thread.data) == b""
        assert mock_log.error.mock_calls == [
            call("Failed to create decryptor: Unable to decrypt cipher INVALID"),
        ]

    @patch("streamlink.stream.hls.hls.log")
    def test_hls_encrypted_missing_uri(self, mock_log: Mock):
        aesKey, aesIv, key = self.gen_key(uri=False)

        self.subject([
            Playlist(0, [key, SegmentEnc(1, aesKey, aesIv)], end=True),
        ])
        self.await_write()

        self.thread.close()
        self.await_close()

        assert b"".join(self.thread.data) == b""
        assert mock_log.error.mock_calls == [
            call("Failed to create decryptor: Missing URI for decryption key"),
        ]

    @patch("streamlink.stream.hls.hls.log")
    def test_hls_encrypted_missing_adapter(self, mock_log: Mock):
        aesKey, aesIv, key = self.gen_key(uri="foo://bar/baz", mock={"exc": InvalidSchema})

        self.subject([
            Playlist(0, [key, SegmentEnc(1, aesKey, aesIv)], end=True),
        ])
        self.await_write()

        self.thread.close()
        self.await_close()

        assert b"".join(self.thread.data) == b""
        assert mock_log.error.mock_calls == [
            call("Failed to create decryptor: Unable to find connection adapter for key URI: foo://bar/baz"),
        ]

    def test_hls_encrypted_aes128(self):
        aesKey, aesIv, key = self.gen_key()
        long = b"Test cipher block chaining mode by using a long bytes string"

        # noinspection PyTypeChecker
        segments = self.subject([
            Playlist(0, [key] + [SegmentEnc(num, aesKey, aesIv) for num in range(4)]),
            Playlist(4, [key] + [SegmentEnc(num, aesKey, aesIv, content=long) for num in range(4, 8)], end=True),
        ])

        self.await_write(3 + 4)
        data = self.await_read(read_all=True)
        self.await_close()

        expected = self.content(segments, prop="content_plain", cond=lambda s: s.num >= 1)
        assert data == expected, "Decrypts the AES-128 identity stream"
        assert self.called(key, once=True), "Downloads encryption key only once"
        assert self.get_mock(key).last_request._request.headers.get("X-FOO") == "BAR"
        assert not any(self.called(s) for s in segments.values() if s.num < 1), "Skips first segment"
        assert all(self.called(s) for s in segments.values() if s.num >= 1), "Downloads all remaining segments"
        assert self.get_mock(segments[1]).last_request._request.headers.get("X-FOO") == "BAR"

    @patch("streamlink.stream.hls.hls.log")
    def test_hls_encrypted_aes128_custom_adapter(self, mock_log: Mock):
        class FooAdapter(requests.adapters.BaseAdapter):
            def close(self):  # pragma: no cover
                return

            def send(self, request, *_, **__):
                res = Response()
                res.request = request
                res.status_code = 200
                res._content = aes_key
                return res

        aes_key, aes_iv, key = self.gen_key(uri="foo://bar/baz", mock={"real_http": True})

        segments = self.subject(
            [
                Playlist(0, [key, SegmentEnc(0, aes_key, aes_iv)], end=True),
            ],
            start=False,
        )

        self.session.http.mount("foo://", FooAdapter())

        self.start()
        self.await_write(1)
        data = self.await_read(read_all=True)
        self.await_close()

        assert data == self.content(segments, prop="content_plain")
        assert mock_log.error.call_args_list == []

    def test_hls_encrypted_aes128_with_map(self):
        aesKey, aesIv, key = self.gen_key()
        map1 = TagMapEnc(1, namespace=self.id(), key=aesKey, iv=aesIv)
        map2 = TagMapEnc(2, namespace=self.id(), key=aesKey, iv=aesIv)
        self.mock("GET", self.url(map1), content=map1.content)
        self.mock("GET", self.url(map2), content=map2.content)

        segments = self.subject([
            Playlist(0, [key, map1] + [SegmentEnc(num, aesKey, aesIv) for num in range(2)]),
            Playlist(2, [key, map2] + [SegmentEnc(num, aesKey, aesIv) for num in range(2, 4)], end=True),
        ])

        self.await_write(1 + 2 + 1 + 2)  # 1 map, 2 segments, 1 map, 2 segments
        data = self.await_read(read_all=True)
        self.await_close()

        assert data == self.content(
            [
                map1,
                segments[0],
                segments[1],
                map2,
                segments[2],
                segments[3],
            ],
            prop="content_plain",
        )

    def test_hls_encrypted_aes128_with_differently_encrypted_map(self):
        aesKey1, aesIv1, key1 = self.gen_key()  # init key
        aesKey2, aesIv2, key2 = self.gen_key()  # media key
        map1 = TagMapEnc(1, namespace=self.id(), key=aesKey1, iv=aesIv1)
        map2 = TagMapEnc(2, namespace=self.id(), key=aesKey1, iv=aesIv1)
        self.mock("GET", self.url(map1), content=map1.content)
        self.mock("GET", self.url(map2), content=map2.content)

        segments = self.subject([
            Playlist(0, [key1, map1, key2] + [SegmentEnc(num, aesKey2, aesIv2) for num in range(2)]),
            Playlist(2, [key1, map2, key2] + [SegmentEnc(num, aesKey2, aesIv2) for num in range(2, 4)], end=True),
        ])

        self.await_write(1 + 2 + 1 + 2)  # 1 map, 2 segments, 1 map, 2 segments
        data = self.await_read(read_all=True)
        self.await_close()

        assert data == self.content(
            [
                map1,
                segments[0],
                segments[1],
                map2,
                segments[2],
                segments[3],
            ],
            prop="content_plain",
        )

    def test_hls_encrypted_aes128_with_plaintext_map(self):
        aesKey, aesIv, key = self.gen_key()
        map1 = TagMap(1, namespace=self.id())
        map2 = TagMap(2, namespace=self.id())
        self.mock("GET", self.url(map1), content=map1.content)
        self.mock("GET", self.url(map2), content=map2.content)

        segments = self.subject([
            Playlist(0, [map1, key] + [SegmentEnc(num, aesKey, aesIv) for num in range(2)]),
            Playlist(2, [map2, key] + [SegmentEnc(num, aesKey, aesIv) for num in range(2, 4)], end=True),
        ])

        self.await_write(1 + 2 + 1 + 2)  # 1 map, 2 segments, 1 map, 2 segments
        data = self.await_read(read_all=True)
        self.await_close()

        assert data == (
            map1.content
            + segments[0].content_plain
            + segments[1].content_plain
            + map2.content
            + segments[2].content_plain
            + segments[3].content_plain
        )

    def test_hls_encrypted_aes128_key_uri_override(self):
        aesKey, aesIv, key = self.gen_key(uri="http://real-mocked/{namespace}/encryption.key?foo=bar")
        aesKeyInvalid = bytes(ord(aesKey[i : i + 1]) ^ 0xFF for i in range(16))
        _, __, key_invalid = self.gen_key(aesKeyInvalid, aesIv, uri="http://mocked/{namespace}/encryption.key?foo=bar")

        # noinspection PyTypeChecker
        segments = self.subject(
            [
                Playlist(0, [key_invalid] + [SegmentEnc(num, aesKey, aesIv) for num in range(4)]),
                Playlist(4, [key_invalid] + [SegmentEnc(num, aesKey, aesIv) for num in range(4, 8)], end=True),
            ],
            options={"hls-segment-key-uri": "{scheme}://real-{netloc}{path}?{query}"},
        )

        self.await_write(3 + 4)
        data = self.await_read(read_all=True)
        self.await_close()

        expected = self.content(segments, prop="content_plain", cond=lambda s: s.num >= 1)
        assert data == expected, "Decrypts stream from custom key"
        assert not self.called(key_invalid), "Skips encryption key"
        assert self.called(key, once=True), "Downloads custom encryption key"
        assert self.get_mock(key).last_request._request.headers.get("X-FOO") == "BAR"

    @patch("streamlink.stream.hls.hls.log")
    def test_hls_encrypted_aes128_incorrect_block_length(self, mock_log: Mock):
        aesKey, aesIv, key = self.gen_key()

        segments = self.subject([
            Playlist(
                0,
                [
                    key,
                    SegmentEnc(0, aesKey, aesIv, append=b"?"),
                    SegmentEnc(1, aesKey, aesIv),
                ],
                end=True,
            ),
        ])
        self.await_write()
        assert self.thread.reader.writer.is_alive()

        self.await_write()
        data = self.await_read(read_all=True)
        self.await_close()

        assert data == self.content([segments[1]], prop="content_plain")
        assert mock_log.error.mock_calls == [
            call("Error while decrypting segment 0: Data must be padded to 16 byte boundary in CBC mode"),
        ]

    @patch("streamlink.stream.hls.hls.log")
    def test_hls_encrypted_aes128_incorrect_padding_length(self, mock_log: Mock):
        aesKey, aesIv, key = self.gen_key()

        padding = b"\x00" * (AES.block_size - len(b"[0]"))
        segments = self.subject([
            Playlist(
                0,
                [
                    key,
                    SegmentEnc(0, aesKey, aesIv, padding=padding),
                    SegmentEnc(1, aesKey, aesIv),
                ],
                end=True,
            ),
        ])
        self.await_write()
        assert self.thread.reader.writer.is_alive()

        self.await_write()
        data = self.await_read(read_all=True)
        self.await_close()

        assert data == self.content([segments[1]], prop="content_plain")
        assert mock_log.error.mock_calls == [call("Error while decrypting segment 0: Padding is incorrect.")]

    @patch("streamlink.stream.hls.hls.log")
    def test_hls_encrypted_aes128_incorrect_padding_content(self, mock_log: Mock):
        aesKey, aesIv, key = self.gen_key()

        padding = (b"\x00" * (AES.block_size - len(b"[0]") - 1)) + bytes([AES.block_size])
        segments = self.subject([
            Playlist(
                0,
                [
                    key,
                    SegmentEnc(0, aesKey, aesIv, padding=padding),
                    SegmentEnc(1, aesKey, aesIv),
                ],
                end=True,
            ),
        ])
        self.await_write()
        assert self.thread.reader.writer.is_alive()

        self.await_write()
        data = self.await_read(read_all=True)
        self.await_close()

        assert data == self.content([segments[1]], prop="content_plain")
        assert mock_log.error.mock_calls == [call("Error while decrypting segment 0: PKCS#7 padding is incorrect.")]

    def test_hls_encrypted_switch_methods(self):
        aesKey1, aesIv1, key_aes128_1 = self.gen_key()
        aesKey2, aesIv2, key_aes128_2 = self.gen_key()
        key_none = Tag("EXT-X-KEY", {"METHOD": "NONE"})

        segments = self.subject([
            Playlist(
                0,
                [
                    key_aes128_1,
                    SegmentEnc(0, aesKey1, aesIv1),
                    SegmentEnc(1, aesKey1, aesIv1),
                    key_none,
                    Segment(2),
                    Segment(3),
                    key_aes128_2,
                    SegmentEnc(4, aesKey2, aesIv2),
                    SegmentEnc(5, aesKey2, aesIv2),
                ],
                end=True,
            ),
        ])

        self.await_write(6)
        data = self.await_read(read_all=True)
        self.await_close()

        expected = self.content(segments, prop="content_plain", cond=lambda s: 0 <= s.num <= 1)
        expected += self.content(segments, cond=lambda s: 2 <= s.num <= 3)
        expected += self.content(segments, prop="content_plain", cond=lambda s: 4 <= s.num <= 5)
        assert data == expected, "Switches between encryption key methods"

    @patch("streamlink.stream.hls.hls.log")
    def test_hls_passthrough_encrypted(self, mock_log: Mock):
        aes_key, aes_iv, key = self.gen_key()
        segments = self.subject(
            options={
                "hls-live-edge": 4,
                "stream-passthrough-encrypted": True,
            },
            playlists=[
                Playlist(0, [key] + [SegmentEnc(num, aes_key, aes_iv) for num in range(4)]),
                Playlist(4, [key] + [SegmentEnc(num, aes_key, aes_iv) for num in range(4, 8)], end=True),
            ],
        )

        self.await_write(8)
        data = self.await_read(read_all=True)
        self.await_close()

        expected = self.content(segments)
        encrypted = self.content(segments, prop="content_plain")
        assert data == expected, "Does not decrypt the stream data"
        assert data != encrypted, "Does not decrypt the stream data"
        assert (
            len([
                call_arg
                for call_arg in mock_log.debug.call_args_list
                if call_arg.args == ("Segments in this playlist are encrypted",)
            ])
            == 1
        )
        assert (
            len([
                call_arg
                for call_arg in mock_log.warning.call_args_list
                if call_arg.args == ("The stream content is encrypted with '%s' and won't be decrypted", "AES-128")
            ])
            == 1
        )


@patch("streamlink.stream.hls.hls.HLSStreamWorker.wait", Mock(return_value=True))
# override default queue polling rate to avoid unnecessary stalling during test runtime
@patch(
    "streamlink.stream.segmented.segmented.SegmentedStreamWriter._queue_get",
    lambda writer: writer._queue.get(block=True, timeout=0.01),
)
class TestHlsInsecureSchemeMedia(TestMixinStreamHLS, unittest.TestCase):
    @patch("streamlink.stream.hls.hls.log")
    def test_http_to_file_segment(self, mock_log: Mock):
        class FileSegment(Segment):
            @property
            def path(self):
                return f"file:///path/{self.num}"

        self.subject([Playlist(0, [Segment(0), FileSegment(1)], end=True)])

        assert self.await_read(read_all=True) == b"", "Rejects the entire playlist"
        assert mock_log.error.call_args_list == [
            call("Prevented access to insecure resource in playlist: base_scheme='http' scheme='file'"),
        ]

    @patch("streamlink.stream.hls.hls.log")
    def test_http_to_file_map(self, mock_log: Mock):
        map0 = TagMap(0, "", attrs={"URI": "file:///path/map0"})
        self.subject([Playlist(0, [map0, Segment(0)], end=True)])

        assert self.await_read(read_all=True) == b"", "Rejects the entire playlist"
        assert mock_log.error.call_args_list == [
            call("Prevented access to insecure resource in playlist: base_scheme='http' scheme='file'"),
        ]

    @patch("streamlink.stream.hls.hls.log")
    def test_http_to_file_key(self, mock_log: Mock):
        key = TagKey(method="AES-128", uri="file:///key", iv=os.urandom(16), keyformat="identity", keyformatversions=1)
        self.subject([Playlist(0, [key, Segment(0)], end=True)])

        assert self.await_read(read_all=True) == b"", "Rejects the entire playlist"
        assert mock_log.error.call_args_list == [
            call("Prevented access to insecure resource in playlist: base_scheme='http' scheme='file'"),
        ]

    @patch("streamlink.stream.hls.hls.log")
    def test_file_to_file(self, mock_log: Mock):
        class FilePlaylist(Playlist):
            def url(self, namespace):
                return "file:///path/playlist"

        class FileSegment(Segment):
            @property
            def path(self):
                return f"file:///path/{self.num}"

            def url(self, namespace):
                return self.path

        segments = self.subject([FilePlaylist(0, [FileSegment(0), FileSegment(1)], end=True)])

        assert self.await_read(read_all=True) == self.content(segments)
        assert mock_log.error.call_args_list == []

    @patch("streamlink.stream.hls.hls.log")
    def test_http_to_https(self, mock_log: Mock):
        class SecureSegment(Segment):
            @property
            def path(self):
                return f"https://mocked/{self.num}"

            def url(self, namespace):
                return self.path

        segments = self.subject([Playlist(0, [SecureSegment(0)], end=True)])

        assert self.await_read(read_all=True) == self.content(segments)
        assert mock_log.error.call_args_list == []

    @patch("streamlink.stream.hls.hls.log")
    def test_https_to_http(self, mock_log: Mock):
        class SecurePlaylist(Playlist):
            def url(self, namespace):
                return f"https://mocked/{namespace}/{self.path}"

        class InsecureSegment(Segment):
            @property
            def path(self):
                return f"http://mocked/{self.num}"

            def url(self, namespace):
                return self.path

        self.subject([SecurePlaylist(0, [InsecureSegment(0)], end=True)])

        assert self.await_read(read_all=True) == b"", "Rejects the entire playlist"
        assert mock_log.error.call_args_list == [
            call("Prevented access to insecure resource in playlist: base_scheme='https' scheme='http'"),
        ]


@pytest.mark.parametrize(
    ("multivariant", "media1", "media2", "raises"),
    [
        pytest.param(
            "http://mocked/multivariant",
            "file:///path/media1",
            "http://mocked/media2",
            pytest.raises(
                OSError,
                match=r".+: Prevented access to insecure resource in playlist: base_scheme='http' scheme='file'$",
            ),
            id="http-to-file-media1",
        ),
        pytest.param(
            "http://mocked/multivariant",
            "http://mocked/media1",
            "file:///path/media2",
            pytest.raises(
                OSError,
                match=r".+: Prevented access to insecure resource in playlist: base_scheme='http' scheme='file'$",
            ),
            id="http-to-file-media2",
        ),
        pytest.param(
            "file:///path/multivariant",
            "file:///path/media1",
            "file:///path/media2",
            does_not_raise,
            id="file-to-file",
        ),
        pytest.param(
            "http://mocked/multivariant",
            "https://mocked/media1",
            "https://mocked/media2",
            does_not_raise,
            id="http-to-https",
        ),
        pytest.param(
            "https://mocked/multivariant",
            "http://mocked/media1",
            "https://mocked/media2",
            pytest.raises(
                OSError,
                match=r".+: Prevented access to insecure resource in playlist: base_scheme='https' scheme='http'$",
            ),
            id="https-to-http-media1",
        ),
        pytest.param(
            "https://mocked/multivariant",
            "https://mocked/media1",
            "http://mocked/media2",
            pytest.raises(
                OSError,
                match=r".+: Prevented access to insecure resource in playlist: base_scheme='https' scheme='http'$",
            ),
            id="https-to-http-media2",
        ),
    ],
)
def test_hls_insecure_scheme_multivariant(
    session: Streamlink,
    requests_mock: rm.Mocker,
    multivariant: str,
    media1: str,
    media2: str,
    raises: nullcontext,
):
    requests_mock.register_uri("GET", media1, text="")
    requests_mock.register_uri("GET", media2, text="")
    requests_mock.register_uri(
        "GET",
        multivariant,
        text="\n".join([
            "#EXTM3U",
            f'#EXT-X-MEDIA:TYPE=AUDIO,GROUP-ID="audio",URI="{media1}",NAME="audio",AUTOSELECT=YES,DEFAULT=YES',
            '#EXT-X-STREAM-INF:BANDWIDTH=1,AUDIO="audio"',
            "#EXT-X-STREAM-INF:BANDWIDTH=1",
            media2,
            "",
        ]),
    )
    with raises:
        HLSStream.parse_variant_playlist(session, multivariant)


@patch("streamlink.stream.hls.hls.HLSStreamWorker.wait", Mock(return_value=True))
class TestHlsReloadTime(TestMixinStreamHLS, unittest.TestCase):
    segments = [
        Segment(0, duration=11),
        Segment(1, duration=7),
        Segment(2, duration=5),
        Segment(3, duration=3),
    ]

    def get_session(self, options=None, *args, **kwargs):
        return super().get_session(
            dict(
                options or {},
                **{
                    "hls-live-edge": 3,
                    "hls-playlist-reload-time": kwargs.pop("reload_time", None),
                },
            ),
        )

    def subject(self, *args, **kwargs):
        kwargs["start"] = False
        super().subject(*args, **kwargs)

        # mock the worker thread's _playlist_reload_time method, so that the main thread can wait on its call
        get_reload_time_called = Event()
        orig_get_reload_time = self.thread.reader.worker._get_reload_time

        def mocked_get_reload_time(*args2, **kwargs2):
            get_reload_time_called.set()
            return orig_get_reload_time(*args2, **kwargs2)

        # immediately kill the writer thread as we don't need it and don't want to wait for its queue polling to end
        def mocked_queue_get():
            return None

        with (
            patch.object(self.thread.reader.worker, "_get_reload_time", side_effect=mocked_get_reload_time),
            patch.object(self.thread.reader.writer, "_queue_get", side_effect=mocked_queue_get),
        ):
            self.start()

            if not get_reload_time_called.wait(timeout=5):  # pragma: no cover
                raise RuntimeError("Missing _get_reload_time() call")

            # close threads first, so that deterministic assertions can be done about the reload time
            self.close()
            self.await_close()

            return self.thread.reader.worker._reload_time

    def test_default(self):
        time = self.subject([Playlist(0, self.segments, end=True, targetduration=4)], reload_time="default")
        assert time == 4, "default sets the reload time to the playlist's target duration"

    def test_segment(self):
        time = self.subject([Playlist(0, self.segments, end=True, targetduration=4)], reload_time="segment")
        assert time == 3, "segment sets the reload time to the playlist's last segment"

    def test_segment_no_segments(self):
        time = self.subject([Playlist(0, [], end=True, targetduration=4)], reload_time="segment")
        assert time == 4, "segment sets the reload time to the targetduration if no segments are available"

    def test_segment_no_segments_no_targetduration(self):
        time = self.subject([Playlist(0, [], end=True, targetduration=0)], reload_time="segment")
        assert time == 6, "sets reload time to 6 seconds when no segments and no targetduration are available"

    def test_live_edge(self):
        time = self.subject([Playlist(0, self.segments, end=True, targetduration=4)], reload_time="live-edge")
        assert time == 8, "live-edge sets the reload time to the sum of the number of segments of the live-edge"

    def test_live_edge_no_segments(self):
        time = self.subject([Playlist(0, [], end=True, targetduration=4)], reload_time="live-edge")
        assert time == 4, "live-edge sets the reload time to the targetduration if no segments are available"

    def test_live_edge_no_segments_no_targetduration(self):
        time = self.subject([Playlist(0, [], end=True, targetduration=0)], reload_time="live-edge")
        assert time == 6, "sets reload time to 6 seconds when no segments and no targetduration are available"

    def test_number(self):
        time = self.subject([Playlist(0, self.segments, end=True, targetduration=4)], reload_time="2")
        assert time == 2, "number values override the reload time"

    def test_number_invalid(self):
        time = self.subject([Playlist(0, self.segments, end=True, targetduration=4)], reload_time="0")
        assert time == 4, "invalid number values set the reload time to the playlist's targetduration"

    def test_no_target_duration(self):
        time = self.subject([Playlist(0, self.segments, end=True, targetduration=0)], reload_time="default")
        assert time == 8, "uses the live-edge sum if the playlist is missing the targetduration data"

    def test_no_data(self):
        time = self.subject([Playlist(0, [], end=True, targetduration=0)], reload_time="default")
        assert time == 6, "sets reload time to 6 seconds when no data is available"


@patch("streamlink.stream.hls.hls.log")
@patch("streamlink.stream.hls.hls.HLSStreamWorker.wait", Mock(return_value=True))
@patch("streamlink.stream.hls.hls.HLSStreamWorker.check_queue_deadline", Mock(return_value=False))
class TestHlsPlaylistParseErrors(TestMixinStreamHLS, unittest.TestCase):
    __stream__ = EventedWriterHLSStream

    class FakePlaylist(NamedTuple):
        is_master: bool = False
        iframes_only: bool = False

    class InvalidPlaylist(Playlist):
        def build(self, *args, **kwargs):
            return "invalid"

    def test_generic(self, mock_log):
        self.subject([self.InvalidPlaylist()])
        assert self.await_read(read_all=True) == b""
        self.await_close()
        assert self.thread.reader.buffer.closed, "Closes the stream on initial playlist parsing error"
        assert mock_log.debug.mock_calls == [call("Reloading playlist")]
        assert mock_log.error.mock_calls == [call("Missing #EXTM3U header")]

    def test_reload(self, mock_log):
        segments = self.subject([
            Playlist(1, [Segment(0)]),
            self.InvalidPlaylist(),
            self.InvalidPlaylist(),
            Playlist(2, [Segment(2)], end=True),
        ])
        self.await_write(2)
        data = self.await_read(read_all=True)
        assert data == self.content(segments)
        self.close()
        self.await_close()
        assert mock_log.warning.mock_calls == [
            call("Reloading failed: Missing #EXTM3U header"),
            call("Reloading failed: Missing #EXTM3U header"),
        ]

    @patch("streamlink.stream.hls.hls.parse_m3u8", Mock(return_value=FakePlaylist(is_master=True)))
    def test_is_master(self, mock_log):
        self.subject([Playlist()])
        assert self.await_read(read_all=True) == b""
        self.await_close()
        assert self.thread.reader.buffer.closed, "Closes the stream on initial playlist parsing error"
        assert mock_log.debug.mock_calls == [call("Reloading playlist")]
        assert mock_log.error.mock_calls == [
            call(f"Attempted to play a variant playlist, use 'hls://{self.stream.url}' instead"),
        ]

    @patch("streamlink.stream.hls.hls.parse_m3u8", Mock(return_value=FakePlaylist(iframes_only=True)))
    def test_iframes_only(self, mock_log):
        self.subject([Playlist()])
        assert self.await_read(read_all=True) == b""
        self.await_close()
        assert self.thread.reader.buffer.closed, "Closes the stream on initial playlist parsing error"
        assert mock_log.debug.mock_calls == [call("Reloading playlist")]
        assert mock_log.error.mock_calls == [call("Streams containing I-frames only are not playable")]


class TestHlsExtAudio:
    @pytest.fixture(autouse=True)
    def _is_usable(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setattr("streamlink.stream.hls.hls.FFMPEGMuxer.is_usable", Mock(return_value=True))

    @pytest.fixture(autouse=True)
    def _playlist(self, request: pytest.FixtureRequest, requests_mock: rm.Mocker):
        params = getattr(request, "param", {})
        with text(params.get("playlist", "hls/test_media_language.m3u8")) as playlist:
            requests_mock.get("http://mocked/path/master.m3u8", text=playlist.read())
            yield

    @pytest.fixture()
    def stream(self, session: Streamlink):
        streams = HLSStream.parse_variant_playlist(session, "http://mocked/path/master.m3u8")
        assert "video" in streams

        return streams["video"]

    def test_no_selection(self, stream: HLSStream):
        assert not isinstance(stream, MuxedHLSStream)
        assert stream.url == "http://mocked/path/playlist.m3u8"

    @pytest.mark.parametrize(
        ("session", "selection"),
        [
            pytest.param({"hls-audio-select": ["EN"]}, "http://mocked/path/en.m3u8", id="English-alpha2"),
            pytest.param({"hls-audio-select": ["ENG"]}, "http://mocked/path/en.m3u8", id="English-alpha3"),
            pytest.param({"hls-audio-select": ["English"]}, "http://mocked/path/en.m3u8", id="English-name"),
            pytest.param({"hls-audio-select": ["English Language"]}, "http://mocked/path/en.m3u8", id="English-name-attr"),
            pytest.param({"hls-audio-select": ["es"]}, "http://mocked/path/es.m3u8", id="Spanish-alpha2"),
            pytest.param({"hls-audio-select": ["spa"]}, "http://mocked/path/es.m3u8", id="Spanish-alpha3"),
            pytest.param({"hls-audio-select": ["spanish"]}, "http://mocked/path/es.m3u8", id="Spanish-name"),
            pytest.param({"hls-audio-select": ["spanish language"]}, "http://mocked/path/es.m3u8", id="Spanish-name-attr"),
        ],
        indirect=["session"],
    )
    def test_selection(self, session: Streamlink, stream: MuxedHLSStream, selection: str):
        assert isinstance(stream, MuxedHLSStream)
        assert [substream.url for substream in stream.substreams] == [
            "http://mocked/path/playlist.m3u8",
            selection,
        ]

    @pytest.mark.parametrize(
        "session",
        [
            pytest.param({"hls-audio-select": ["en", "es"]}, id="alpha2"),
            pytest.param({"hls-audio-select": ["eng", "spa"]}, id="alpha3"),
            pytest.param({"hls-audio-select": ["English", "Spanish"]}, id="name"),
            pytest.param({"hls-audio-select": ["English language", "Spanish language"]}, id="name-attr"),
        ],
        indirect=["session"],
    )
    def test_multiple(self, session: Streamlink, stream: MuxedHLSStream):
        assert isinstance(stream, MuxedHLSStream)
        assert [substream.url for substream in stream.substreams] == [
            "http://mocked/path/playlist.m3u8",
            "http://mocked/path/en.m3u8",
            "http://mocked/path/es.m3u8",
        ]

    @pytest.mark.parametrize(
        "session",
        [
            pytest.param({"hls-audio-select": ["*"]}, id="wildcard"),
        ],
        indirect=["session"],
    )
    def test_wildcard(self, session: Streamlink, stream: MuxedHLSStream):
        assert isinstance(stream, MuxedHLSStream)
        assert [substream.url for substream in stream.substreams] == [
            "http://mocked/path/playlist.m3u8",
            "http://mocked/path/en.m3u8",
            "http://mocked/path/es.m3u8",
            "http://mocked/path/de.m3u8",
        ]

    @pytest.mark.parametrize(
        ("session", "_playlist", "expected"),
        [
            pytest.param(
                {"hls-audio-select": ["und", "qaa", "en-x-foo", "invalid"]},
                {"playlist": "hls/test_media_language_special.m3u8"},
                [
                    "http://mocked/path/playlist.m3u8",
                    "http://mocked/path/qaa.m3u8",
                    "http://mocked/path/en-x-foo.m3u8",
                    "http://mocked/path/invalid.m3u8",
                    "http://mocked/path/und.m3u8",
                ],
                id="special-reserved-private-invalid",
            ),
            pytest.param(
                {"hls-audio-select": ["Undetermined", "Reserved", "Private use", "does NOT exist"]},
                {"playlist": "hls/test_media_language_special.m3u8"},
                [
                    "http://mocked/path/playlist.m3u8",
                    "http://mocked/path/qaa.m3u8",
                    "http://mocked/path/en-x-foo.m3u8",
                    "http://mocked/path/invalid.m3u8",
                    "http://mocked/path/und.m3u8",
                ],
                id="name-attribute",
            ),
            pytest.param(
                {"hls-audio-select": ["*"]},
                {"playlist": "hls/test_media_language_special.m3u8"},
                [
                    "http://mocked/path/playlist.m3u8",
                    "http://mocked/path/qaa.m3u8",
                    "http://mocked/path/en-x-foo.m3u8",
                    "http://mocked/path/invalid.m3u8",
                    "http://mocked/path/und.m3u8",
                    "http://mocked/path/qqq.m3u8",
                ],
                id="all",
            ),
        ],
        indirect=["session", "_playlist"],
    )
    def test_parse_media_language(
        self,
        caplog: pytest.LogCaptureFixture,
        session: Streamlink,
        stream: MuxedHLSStream,
        _playlist: None,
        expected: list[str],
    ):
        assert isinstance(stream, MuxedHLSStream)
        assert [substream.url for substream in stream.substreams] == expected
        assert [
            record.message
            for record in caplog.get_records(when="setup")
            if record.name == "streamlink.stream.hls" and record.levelno == logging.WARNING
        ] == [
            "Unrecognized language for media playlist: language='invalid' name='Does not exist'",
        ]

    @pytest.mark.parametrize(
        "session",
        [
            pytest.param({"hls-audio-select": ["*"]}, id="names"),
        ],
        indirect=["session"],
    )
    def test_substream_names(self, session: Streamlink, stream: MuxedHLSStream):
        assert isinstance(stream, MuxedHLSStream)
        assert [substream.name for substream in stream.substreams] == [
            None,
            "audio",
            "audio",
            "audio",
        ]


class TestM3U8ParserLogging:
    @pytest.mark.parametrize(("loglevel", "has_logs"), [("trace", False), ("all", True)])
    def test_log(self, caplog: pytest.LogCaptureFixture, loglevel: str, has_logs: bool):
        caplog.set_level(loglevel, "streamlink")

        parser = M3U8Parser[M3U8, HLSSegment, HLSPlaylist]()
        with text("hls/test_1.m3u8") as pl:
            data = pl.read()
        parser.parse(data)

        assert bool(caplog.records) is has_logs



from streamlink.plugins.pandalive import Pandalive
from tests.plugins import PluginCanHandleUrl


class TestPluginCanHandleUrlPandalive(PluginCanHandleUrl):
    __plugin__ = Pandalive

    should_match_groups = [
        ("https://www.pandalive.co.kr/live/play/CHANNEL", {"channel": "CHANNEL"}),
        ("https://w2.pandalive.co.kr/en/play/CHANNEL", {"channel": "CHANNEL"}),
    ]



import pytest

from streamlink.utils.data import search_dict


@pytest.mark.parametrize(
    ("args", "expected"),
    [
        ((["one", "two"], "one"), []),
        (({"two": "test2"}, "one"), []),
        (({"one": "test1", "two": "test2"}, "one"), ["test1"]),
        (({"one": {"inner": "test1"}, "two": "test2"}, "inner"), ["test1"]),
        (({"one": [{"inner": "test1"}], "two": "test2"}, "inner"), ["test1"]),
        (({"one": [{"inner": "test1"}], "two": {"inner": "test2"}}, "inner"), ["test1", "test2"]),
    ],
)
def test_search_dict(args, expected):
    assert list(search_dict(*args)) == expected
